import operator
import numpy as np
import numba
from numba.core.dispatcher import Dispatcher
from numba import types, njit, f8, i8, u8, i4, u1, u2, u4, literally, generated_jit, boolean, literal_unroll
from numba.typed import List, Dict
from numba.types import ListType, DictType, unicode_type, void, Tuple, UniTuple, StructRef
from numba.experimental import structref
from numba.experimental.structref import new, define_boxing, define_attributes, _Utils, StructRefProxy
from numba.extending import lower_cast, NativeValue, box, unbox, overload_method, intrinsic, overload_attribute, intrinsic, lower_getattr_generic, overload, SentryLiteralArgs
from numba.core.typing.templates import AttributeTemplate
from numba.core.errors import NumbaError, NumbaPerformanceWarning
from cre.caching import gen_import_str, unique_hash_v, import_from_cached, source_to_cache, source_in_cache, cache_safe_exec, get_cache_path
from cre.context import cre_context
from cre.structref import define_structref, define_structref_template, StructRefType
from cre.utils import (cast, _sizeof_type, _nullify_attr, new_w_del, _memcpy, _func_from_address, decode_idrec, lower_getattr,  
                       _decref_ptr, _incref_ptr, _incref_structref, _decref_structref, _ptr_from_struct_incref, ptr_t, _load_ptr,
                       _obj_cast_codegen)
from cre.utils import PrintElapse, encode_idrec, assign_to_alias_in_parent_frame, as_typed_list, lower_setattr, _store, _store_safe, _tuple_getitem
from cre.vector import VectorType
from cre.fact import Fact, gen_fact_import_str, get_offsets_from_member_types
from cre.var import Var, var_extend,  VarType, VarTypeClass
from cre.obj import CREObjType, cre_obj_field_dict, CREObjTypeClass, CREObjProxy, member_info_type, set_chr_mbrs, cre_obj_get_item_t_id_ptr, cre_obj_set_item, cre_obj_get_item, PRIMITIVE_MBR_ID
from cre.core import T_ID_FUNC, T_ID_STR, register_global_default
from cre.make_source import make_source, gen_def_func, gen_assign, gen_if, gen_not, resolve_template, gen_def_class
from numba.core import imputils, cgutils
from numba.core.datamodel import default_manager, models, register_default
from numba.experimental.function_type import _get_wrapper_address

# Import to ensure that .format is defined
import cre.type_conv


from operator import itemgetter
from copy import copy
from os import getenv
import inspect, cloudpickle, pickle
from textwrap import dedent, indent
from collections.abc import Iterable
import warnings

from llvmlite import binding as ll
from llvmlite import ir
import cre_cfuncs

from numba.extending import intrinsic, overload_method, overload

def all_args_are_const(arg_types):
    '''Helper function for checking if a set of arg_types are all primative types'''
    all_const = True
    for at in arg_types:
        if(isinstance(at, VarTypeClass) or isinstance(at, CREFuncTypeClass)):
            all_const = False
            break
    return all_const

def dispatcher_sig_from_arg_types(disp, arg_types):
    cres = disp.overloads.get(arg_types,None)
    if(cres is not None):
        sig = cres.signature
    else:
        # Note: get_call_template is an internal method of numba.Dispatcher 
        (template,*rest) = disp.get_call_template(arg_types,{})
        #Note: this can be finicky might need to find the best one
        sig = [x for x in template.cases if x.args==arg_types][0]
    return sig

# ----------------------------------------------------------------------
# : Record Array Struct definitions

# Struct that indicates a span of 'head_infos' that are all 
#  associated with the same variable
_head_range_type = np.dtype([
    ('start', np.uint16), # Start of span
    ('end', np.uint16) # Start + length
])
head_range_type = numba.from_dtype(_head_range_type)

# ------------------------------
# ARGINFO type enums
ARGINFO_CONST = u1(1)
ARGINFO_VAR = u1(2)
ARGINFO_FUNC = u1(3)
ARGINFO_FUNC_UNEXPANDED = u1(4)

# ------------------------------
# STATUS enums
CFSTATUS_FALSEY = u1(0)
CFSTATUS_TRUTHY = u1(1)
CFSTATUS_NULL_DEREF = u1(2)
CFSTATUS_ERROR = u1(3)

# ------------------------------
# REFKIND enums
REFKIND_PRIMATIVE = u4(1)
REFKIND_UNICODE = u4(2)
REFKIND_STRUCTREF = u4(3)

# Struct for each argument to a CREFunc which might be assigned 
#  to a primative constant, Var, or other CREFunc.
_arg_infos_type = np.dtype([
    # Enum for CONST, VAR, FUNC, FUNC_UNEXPANDED.
    ('type', np.uint16), 
    # t_id of the argument
    ('t_id', np.uint16), 
    # Whether is Var w/ dereference (e.g. V.nxt.nxt).
    ('has_deref', np.int32), 
    # Ptr to Var or CREFUnc.
    ('ptr', np.int64) 
]) 
arg_infos_type = numba.from_dtype(_arg_infos_type)

# Struct w/ instructions for executing a child CREFunc that needs to be 
#  run before the main root CREFunc in a composition of CREFuncs
_instr_type = np.dtype([
    # The ptr to the CREFunc
    ('cf_ptr', np.int64), 
    # Data ptr to copy the CREFunc's return value into.
    ('return_data_ptr', np.int64), 
    # The byte-width of the return value  
    ('size',np.uint32), 
    # 1 for unicode_type 2 for other. Needed because unicode_strings 
    #  don't have meminfo ptrs as first member.
    ('ref_kind',np.uint32)
])
instr_type = numba.from_dtype(_instr_type)


# Struct for head_infos 
_head_info_type = np.dtype([
    # The pointer to the CREFunc for which this is a head argument.
    ('cf_ptr', np.int64),
    # Enum for CONST, VAR, FUNC, FUNC_UNEXPANDED.
    ('type', np.uint8),
    # If the Var associated with this has a dereference chain.
    ('has_deref', np.uint8),
    # The index of the root argument this head corresponds to.
    ('arg_ind', np.uint16),
    # t_id for the base value's type.
    ('base_t_id', np.uint16),
    # t_id for the head value's type.
    ('head_t_id', np.uint16),
    # The pointer to the Var for this head. Its ref is held in "ref{i}" member.
    ('var_ptr', np.int64),
    # Data pointer for the a{i} member for the base.
    ('arg_data_ptr', np.int64),
    # Data pointer for the h{i} member for the resolved head.
    ('head_data_ptr', np.int64),
    # 1 for unicode_type 2 for other. Needed because unicode_strings don't 
    #  have meminfo ptrs as first member so need different inc/decref implmentation.
    ('ref_kind', np.uint32),
    # The byte-width of the head value.
    ('head_size', np.uint32)])
head_info_type = numba.from_dtype(_head_info_type)

# -----------------------------------------------------------------
# : CREFunc_method decorator

def warn_cant_compile(func_name, cre_f_name, e):
    s = f'''
########## WARNING CAN'T COMPILE {cre_f_name}.{func_name}() ########### 
{indent(str(e),'  ')} 
########################################################
Numba was unable to compile {func_name}() for CREFunc {cre_f_name}. 
Using objmode (i.e. native python) instead. To ignore warning set nopython=False.\n\
'''
    warnings.warn(s, NumbaPerformanceWarning)


def CREFunc_assign_method_addr(cf_type, fn_name, addr):
    name = f"{cf_type.symbol_prefix}_{fn_name}"
    ll.add_symbol(name, addr)

def CREFunc_method(cf_type, sig, _fn_name=None, on_error='error',**kwargs):
    '''A decorator used in CREFunc codegens for trying to compile member
         methods like 'call_heads' and registering their ptrs as globals.
    '''
    def wrapper(func):        
        if(isinstance(func,Dispatcher)):
            dispatcher = func
            fn_name = func.py_func.__name__ if _fn_name is None else _fn_name 
        else:
            fn_name = func.__name__ if _fn_name is None else _fn_name 
            # Try to compile the CREFunc
            try:
                dispatcher = njit(sig, cache=True, **kwargs)(func)
            except NumbaError as e:
                if(on_error == "error"): raise e
                if(on_error == "warn"):
                    warn_cant_compile(fn_name, cf_type.func_name, e)
                return None
        addr = _get_wrapper_address(dispatcher, sig)
        CREFunc_assign_method_addr(cf_type, fn_name, addr)
        # dispatcher.cre_method_addr = addr
        if(not hasattr(cf_type,'dispatchers')):
            cf_type.dispatchers = {}    
        cf_type.dispatchers[fn_name] = dispatcher
        return dispatcher
    return wrapper

# ----------------------------------------------------------------------
# : CREFunc jitted side

OriginData, OriginDataType = define_structref("OriginData", {
    "ptr" : i8,
    "name" : unicode_type,
    "expr_template" : unicode_type,
    "shorthand_template" : unicode_type,
    "repr_const_addrs" : i8[::1],
})

cre_func_fields_dict = {
    **cre_obj_field_dict, 

    # Keeps track of data pertaining to the base CREFunc that originated this 
    # CREFunc including a weak pointer, it's name and various ways of printing it.
    "origin_data" : OriginDataType,

    # The number of arguments taken by this cf.
    "n_args" : i8,

    # The number of arguments taken by the original call() of this cf.
    "n_root_args" : i8,

    # The args to the root CREFunc 
    "root_arg_infos" : arg_infos_type[::1],

    # Maps base arguments to particular head_infos
    "head_ranges" : head_range_type[::1],

    # Keeps references to each head variable, their types, the CREFunc they live in 
    #  among other things
    "head_infos" : head_info_type[::1],

    # A set of instructions for calling CREFuncs that need to be called as
    #  prerequisites to this one, as part of a function composition.
    "prereq_instrs" : instr_type[::1],

    # Data ptr of the return value    
    "return_data_ptr" : i8,
        
    # The address for the root CREFunc's 'call_head' implementation.
    #  Essentially the CREFunc's main call function.     
    "call_heads_addr" : i8,

    # The address for the root CREFunc's 'call_self' implementation.
    #  Calls call_heads on any values stored in 'h{i}' (i.e. 'head') slots.
    "call_self_addr" : i8,
    
    # The address for the root CREFunc's 'resolve_heads_addr' implementation.
    #  Dereferences any objects in 'a{i}' (i.e. 'arg') slots and writes them
    #  do each corresponding 'h{i}' (i.e. 'head') slot.
    "resolve_heads_addr" : i8,

    # True if the func has beed initialized
    "is_initialized" : types.boolean,

    # True if this func is a ptr func
    "is_ptr_func" : types.boolean,

    # NOTE: Not Used 
    # True if dereferencing the head args succeeded  
    # "heads_derefs_succeeded" : types.boolean,

    # The composition depth
    "depth" : types.uint16,

    # NOTE: Not Used 
    # True if check in this and all children suceeded
    # "exec_passed_checks" : types.boolean,
    
    "return_t_id" : u2,

    "has_any_derefs": u1,

    "is_composed" : u1,

    # Other members like a0,a1,... h0,h1,... etc. filled in on specialize
}

@structref.register
class CREFuncTypeClass(types.Callable, CREObjTypeClass):
    '''The numba type class for CREFuncs. '''
    t_id = T_ID_FUNC
    type_cache = {}

    def __new__(cls, return_type=None, arg_types=None, is_composed=False, name=None, long_hash=None):
        if(name is None): name = "CREFuncType"
        if(isinstance(arg_types,list)): arg_types = tuple(arg_types)

        # Check if this type is already cached 
        unq_tup = (return_type, arg_types, is_composed, name, long_hash)
        self = cls.type_cache.get(unq_tup,None)
        if(self is not None):
            return self

        self = super().__new__(cls)

        # If full signature isn't specified then use unspecialized struture 
        if(return_type is None or arg_types is None):
            field_dict = {**cre_func_fields_dict}

        # Otherwise define type-specific slots for 
        #    a{i}: Slot for base CREObject like arguments that will be  
        #            dereferenced into heads (i.e. obj.nxt.nxt.val)
        #    h{i}: Slot for type type-specific head arguments
        #    ref{i} : Slot for Var or CREFunc instances that are part of composition
        #    return_val: Slot for return value.
        #    chr_mbrs_infos: Characteristic Member Infos -- see CREObject
        else:
            arg_fields = {}
            for i,t in enumerate(arg_types):
                arg_fields[f'a{i}'] = CREObjType
                arg_fields[f'h{i}'] = t

            N_MINFOS = 1+len(arg_types)*3

            field_dict = {**cre_func_fields_dict,
                'return_val' : return_type,
                **arg_fields,
                **{f'ref{i}' : CREObjType for i in range(len(arg_types))},
                'chr_mbrs_infos' : UniTuple(member_info_type,N_MINFOS),
            }

        self.func_name = name
        self.return_type = return_type
        self.arg_types = arg_types
        self.is_composed = is_composed
        self.t_id = T_ID_FUNC
        self.long_hash = long_hash
        self._field_dict = field_dict
        cls.type_cache[unq_tup] = self

        # Apply typical content of __init__() here to avoid reinitialization   
        types.StructRef.__init__(self,[(k,v) for k,v in field_dict.items()])
        self.name = repr(self)
        return self

    def __init__(self,*args,**kwargs):
        pass

    # def is_precise(self):
    #     return (self.return_type is not None and 
    #             self.arg_types is not None and
    #             self.is_composed is not None)

    # -------------------------------------------
    # Impl these 3 funcs to subclass types.Callable so can @overload_method('__call__')
    def get_call_type(self, context, args, kws):
        if(all_args_are_const(args)):
            return self.return_type(*args)
        else:
            ty = CREFuncTypeClass(return_type=self.return_type,name="FuncComp")
            return ty(*args)

    def get_call_signatures(self):
        return [self.return_type(*self.arg_types),
                CREFuncType(*([CREFuncType]*len(self.arg_types))),
                CREFuncType(*([VarType]*len(self.arg_types)))
                ]

    def get_impl_key(self, sig):
        return (type(self), '__call__')
    # -------------------------------------------

    def __str__(self):
        if(self.return_type is not None):
            arg_s = str(self.arg_types) if self.arg_types is not None else "(...)"
            return f"{self.func_name}{arg_s}->{self.return_type}{':c' if self.is_composed else ''}"
        else:
            return f"{self.func_name}"

    def __repr__(self):
        if(self.return_type is not None):
            hsh = f"_{self.long_hash}" if self.long_hash is not None else ""
            arg_s = str(self.arg_types) if self.arg_types is not None else "(...)"
            return f"CREFunc[{self.func_name}{hsh}{arg_s}->{self.return_type}{':c' if self.is_composed else ''}]"
        else:
            return f"{self.func_name}"


    @property
    def symbol_prefix(self):
        if(not hasattr(self, "_symbol_prefix")):
            if(self.func_name == "CREFuncType"):
                self._symbol_preix = "CREFuncType"
            else:
                shortened_hash = self.long_hash[:18]
                self._symbol_preix = f"CREFunc_{self.func_name}_{shortened_hash}"
        return self._symbol_preix

    def __getstate__(self):
        # Need to delete any functions members to avoid loading undefined functions
        #  when the type is unpickled. Otherwise "FunctionType not have '_key'" error.
        d = self.__dict__.copy()
        if('call_heads' in d): del d['call_heads']
        if('check' in d): del d['check']
        if('ctor' in d): del d['ctor']
        if('_code' in d): del d['_code']
        if('dispatchers' in d): del d['dispatchers']
        return d


CREFuncType = CREFuncTypeClass()
register_global_default("CREFunc", CREFuncType)

@lower_cast(CREFuncTypeClass, CREFuncType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)

# -----------------------------------------------------------------------
# : CREFunc Proxy Class

def new_cre_func(name, members):
    ''' Returns a singleton instance for a CREFunc definition.'''
    call_func = members['call']

    cf_type = define_CREFunc(name, members)
    if(not hasattr(members, 'arg_types')): 
        members['arg_types'] = members['signature'].args
    if(not hasattr(members, 'return_type')): 
        members['return_type'] = members['signature'].return_type
    n_args = len(members['arg_types'])

    expr_template = f"{name}({', '.join([f'{{{i}}}' for i in range(n_args)])})"
    shorthand_template = members.get('shorthand',expr_template)

    cre_func = cf_type.ctor(
        name,
        expr_template,
        shorthand_template                
    )

    cre_func._return_type = members['return_type']
    cre_func._arg_types = members['arg_types']
    _vars = []
    get_return_val_impl(cre_func.return_type)
    for i, (arg_alias, typ) in enumerate(zip(cf_type.root_arg_names, cf_type.arg_types)):
        v = Var(typ,arg_alias); _vars.append(v);
        set_var_arg(cre_func, i, v)
        set_base_arg_val_impl(typ)
        set_const_arg_impl(typ)
    reinitialize(cre_func)

    cre_func._type = cf_type

    # This prevent _vars from freeing until after reinitialize
    _vars = None 
    return cre_func

def _standardize_py_args(self, py_args):
    '''Helper function that: 
        1) Determines if all arguments are constants.
        2) Replaces type instances with Var instances.
        3) Resolves the head types of the arguments.
    '''
    head_types = [] 
    py_args = list(py_args)
    all_const = True
    for i, x in enumerate(py_args):
        # If given a type make into a Var
        if(isinstance(x, types.Type)):
            x = py_args[i] = Var(x, self.root_arg_names[i])
        if(isinstance(x, Var)):
            head_types.append(x.head_type)
            all_const = False
        elif(isinstance(x, CREFunc)):
            head_types.append(x.return_type)
            all_const = False
        elif(isinstance(x, UntypedCREFunc)):
            raise ValueError("Cannot compose UntypedCREFunc.")
        else:
            head_types.append(resolve_return_type(x))
    head_types = tuple(head_types)
    return py_args, all_const, head_types


class CREFunc(StructRefProxy):

    def __new__(self,*args, **kwargs):
        ''' A decorator function that builds a new CREFunc'''
        if(len(args) > 1): raise ValueError("CREFunc() takes at most one position argument 'signature'.")
        
        def wrapper(call_func):
            # Make new cre_func_type
            members = kwargs
            members['call'] = call_func
            members['root_arg_names'] = inspect.getfullargspec(call_func)[0]
            name = call_func.__name__

            # If no signature is given then create an untyped instance that can be specialized later
            if('signature' not in members):
                return UntypedCREFunc(name,members)

            # Otherwise define a CREFunc with 'name' and 'members' and return
            #  a new singleton instance of it.
            return new_cre_func(name, members)            

        if(len(args) == 1):
            if(isinstance(args[0],(str, numba.core.typing.templates.Signature))):
                kwargs['signature'] = args[0]
            elif(hasattr(args[0],'__call__')):
                return wrapper(args[0])
            else:
                raise ValueError(f"Unrecognized type {type(args[0])} for 'signature'")        

        return wrapper


    def __call__(self,*args):
        if(len(args) != self.n_args):
            raise ValueError(f"Got {len(args)} args for {str(self)} with {self.n_args} positional arguments.")
        self._ensure_has_types()
        arg, all_const, head_type = _standardize_py_args(self, args)
        
        # Call Case, will return a value
        if(all_const):            
            # Assign each argument to it's slot h{i} in the func's memory space
            for i, arg in enumerate(args):
                if(isinstance(arg, CREObjProxy)):
                    impl = set_base_arg_val_impl(CREObjType)
                else:
                    impl = set_base_arg_val_impl(self._arg_types[i])
                impl(self, i, arg)

            
            status = cre_func_resolve_call_self(self)

            if(status == CFSTATUS_NULL_DEREF):
                raise Exception(f"{str(self)} failed to execute because of a dereferencing error.")
            elif(status == CFSTATUS_ERROR):
                raise Exception(f"{str(self)} failed to execute because of an internal error.")

            args = None
            get_ret = get_return_val_impl(self._return_type)
            return get_ret(self)

        # Compose Case, will return a new CREFunc
        else:
            new_cf = cf_deep_copy_ep(self)
            # new_cf = self
            new_cf._return_type = self.return_type
            for i, arg in enumerate(args):
                # Optmization: Assign arguments via entry_points
                #  to skip numba's internal type checking.  
                if(isinstance(arg, Var)):
                    set_var_arg_ep(new_cf, i, arg)
                elif(isinstance(arg, CREFunc)):
                    set_func_arg_ep(new_cf, i, arg)
                else:
                    # print(arg, self.arg_types[i])
                    impl = set_const_arg_impl(arg, use_ep=True)
                    impl(new_cf, i, arg)

            reinitialize(new_cf)

            # NOTE: Slowest step because of cre_context()
            new_cf.recover_arg_types()

            name, long_hash = None, None
            if(hasattr(self,'_type')):
                func_name = self._type.func_name
                long_hash = self._type.long_hash
            new_cf._type = CREFuncTypeClass(self._return_type, self._arg_types,
                                is_composed=new_cf.is_composed,
                                name=func_name, long_hash=long_hash)
            args = None
            return new_cf

    def __str__(self):
        return cre_func_str(self, True, True, "")

    def __repr__(self):
        return cre_func_str(self, False, True, "")

    def __hash__(self):
        from cre.dynamic_exec import cre_func_hash
        return cre_func_hash(self)

    def __copy__(self):
        return cre_func_copy(self)

    def copy(self):
        return cre_func_copy(self)

    @property
    def unique_str(self):
        return cre_func_unique_string(self)

    def minimal_str(self, show_derefs=False, ignore_funcs=[]):
        ignore_pattern = "".join([f"{f.func_name}," for f in ignore_funcs]) 
        return cre_func_str(self, True, show_derefs, ignore_pattern)

    @property
    def return_type(self):
        self._ensure_has_types()        
        return self._return_type

    @property
    @njit(cache=True)
    def return_t_id(self):
        return self.return_t_id

    @property
    def arg_types(self):
        self._ensure_has_types()        
        return self._arg_types

    @property
    def root_arg_names(self):
        return self._type.root_arg_names

    @property
    def signature(self):
        self._ensure_has_types()        
        return self._return_type(*self._arg_types)

    @property
    def precise_type(self):
        if(self._type.return_type is not None and
           self._type.arg_types is not None):
            return self._type
        else:
            return CREFuncTypeClass(
            self.return_type, self.arg_types,
            is_composed=self.is_composed)

    @property
    def func_name(self):
        return self._type.func_name

    @property
    def long_hash(self):
        return self._type.long_hash

    @property
    def right_commutes(self):
        return getattr(self._type,'right_commutes',{})

    @property
    def commutes(self):
        return self._type.commutes

    def _ensure_has_types(self):
        '''Ensures that the CREFunc has a _retur_type and _arg_types 
            which it might be missing if instantied in jitted code.
        '''
        if(getattr(self, '_return_type',None) is None):
            return_type = getattr(self, '_type', CREFuncType).return_type
            if(return_type is not None):
                self._return_type = return_type
                if(self._type.arg_types is None):
                    self.recover_arg_types()
                else:
                    self._arg_types = self._type.arg_types
            else:
                self.recover_return_type()
                self.recover_arg_types()

    def set_var_arg(self, i, val):
        set_var_arg(self, i, val)

    def set_func_arg(self, i, val):
        set_func_arg(self, i, val)

    def set_const_arg(self, i, val):
        impl = set_const_arg_impl(arg)
        impl(new_cf, i, arg)

    def get_const_root_arg(self, i, _typ=None):
        impl = get_const_root_arg_impl(_typ)
        return impl(self, i)




    @property
    @njit(i8(CREFuncType), cache=True)
    def n_args(self):
        return self.n_args

    @property
    @njit(i8(CREFuncType), cache=True)
    def n_funcs(self):
        return len(self.prereq_instrs)+1

    @property
    @njit(head_range_type[::1](CREFuncType), cache=True)
    def head_ranges(self):
        return self.head_ranges

    @property
    @njit(head_info_type[::1](CREFuncType), cache=True)
    def head_infos(self):
        return self.head_infos

    @property
    @njit(u1(CREFuncType), cache=True)
    def is_composed(self):
        return self.is_composed

    @property
    @njit(i8(CREFuncType), cache=True)
    def depth(self):
        return i8(self.depth)

    @property
    @njit(unicode_type(CREFuncType), cache=True)
    def name(self):
        return self.origin_data.name


    def recover_return_type(self):
        self._return_type = cre_context().get_type(t_id=self.return_t_id, ensure_retro=False)

    def recover_arg_types(self):
        context = cre_context()
        arg_types = []
        for t_id in get_base_t_ids_ep(self):
            arg_types.append(context.get_type(t_id=t_id))

        self._arg_types = arg_types

    @property    
    def base_var_ptrs(self):
        return get_base_var_ptrs(self)  

    @property    
    def head_var_ptrs(self):
        return get_head_var_ptrs(self)    

    @property    
    def base_vars(self):
        base_vars = []
        for var_ptr in self.base_var_ptrs:
            base_vars.append(Var.from_ptr(var_ptr))
        return base_vars

    @property
    def head_vars(self):
        return get_head_vars(self)

    @property    
    def base_var_aliases(self):
        pattern = get_base_var_aliases_pattern(self)
        return pattern.split(",")

    @property
    def vars(self):
        return self.base_vars

    @property    
    def base_t_ids(self):
        return get_base_t_ids_ep(self)

    @property    
    def arg_t_ids(self):
        return get_base_t_ids_ep(self)

    

    def get_ptr(self):
        return func_get_ptr(self)
    @classmethod
    def from_ptr(cls, ptr):
        return func_from_ptr(ptr)

    def __lt__(self, other): 
        from cre.default_funcs import LessThan
        return LessThan(self, other)
    def __le__(self, other): 
        from cre.default_funcs import LessThanEq
        return LessThanEq(self, other)
    def __gt__(self, other): 
        from cre.default_funcs import GreaterThan
        return GreaterThan(self, other)
    def __ge__(self, other):
        from cre.default_funcs import GreaterThanEq
        return GreaterThanEq(self, other)
    def __eq__(self, other): 
        from cre.default_funcs import Equals
        return Equals(self, other)
    def __ne__(self, other): 
        from cre.default_funcs import Equals
        return ~Equals(self, other)

    def __add__(self, other):
        from cre.default_funcs import Add
        return Add(self, other)

    def __radd__(self, other):
        from cre.default_funcs import Add
        return Add(other, self)

    def __sub__(self, other):
        from cre.default_funcs import Subtract
        return Subtract(self, other)

    def __rsub__(self, other):
        from cre.default_funcs import Subtract
        return Subtract(other, self)

    def __mul__(self, other):
        from cre.default_funcs import Multiply
        return Multiply(self, other)

    def __rmul__(self, other):
        from cre.default_funcs import Multiply
        return Multiply(other, self)

    def __truediv__(self, other):
        from cre.default_funcs import Divide
        return Divide(self, other)

    def __rtruediv__(self, other):
        from cre.default_funcs import Divide
        return Divide(other, self)

    def __floordiv__(self, other):
        from cre.default_funcs import FloorDivide
        return FloorDivide(self, other)

    def __rfloordiv__(self, other):
        from cre.default_funcs import FloorDivide
        return FloorDivide(other, self)

    def __pow__(self, other):
        from cre.default_funcs import Power
        return Power(self, other)

    def __rpow__(self, other):
        from cre.default_funcs import Power
        return Power(other, self)

    def __mod__(self, other):
        from cre.default_funcs import Modulus
        return Modulus(other, self)

    def __and__(self,other):
        from cre.conditions import cre_func_to_cond, conditions_and
        self = cre_func_to_cond(self)
        if(isinstance(other, CREFunc)): other = cre_func_to_cond(other)
        # print("<<", type(self), type(other))
        return conditions_and(self, other)

    def __or__(self,other):
        from cre.conditions import cre_func_to_cond, conditions_or
        self = cre_func_to_cond(self)
        if(isinstance(other, CREFunc)): other = cre_func_to_cond(other)
        return conditions_or(self, other)

    def __invert__(self):
        from cre.conditions import literal_ctor, literal_to_cond, literal_not
        return literal_to_cond(literal_not(literal_ctor(self)))


define_boxing(CREFuncTypeClass, CREFunc)


# --------------------------
# : Getters for various attribute definitions 

@njit(i8[::1](CREFuncType), cache=True)
def get_base_var_ptrs(self):
    base_var_ptrs = np.empty(self.n_args,dtype=np.int64)
    for i, hrng in enumerate(self.head_ranges):
        hi = self.head_infos[hrng.start]
        v = cast(hi.var_ptr, VarType)
        base_var_ptrs[i] = v.base_ptr
    return base_var_ptrs

@njit(cache=True)
def get_base_var_aliases_pattern(self):
    pattern = ""
    for var_ptr in self.base_var_ptrs:
        v = cast(var_ptr, VarType)
        pattern += f"{v.alias},"
    if(len(pattern) > 0):
        pattern = pattern[:-1]
    return pattern
 

@overload_attribute(CREFuncTypeClass, "base_var_ptrs")
def overload_base_var_ptrs(self):
    return get_base_var_ptrs.py_func

@njit(u2[::1](CREFuncType), cache=True)
def get_base_t_ids(self):
    t_ids = np.empty(self.n_args,dtype=np.uint16)
    for i, hrng in enumerate(self.head_ranges):
        hi = self.head_infos[hrng.start]
        t_ids[i] = hi.base_t_id
    return t_ids

get_base_t_ids_ep = get_base_t_ids.overloads[(CREFuncType,)].entry_point

@overload_attribute(CREFuncTypeClass, "base_t_ids")
def overload_base_t_ids(self):
    return get_base_t_ids.py_func


var_lst_t = ListType(VarType)
@njit(ListType(ListType(VarType))(CREFuncType), cache=True)
def get_head_vars(self):
    head_vars = List.empty_list(var_lst_t)
    for i, hrng in enumerate(self.head_ranges):
        base_heads = List.empty_list(VarType)
        for j in range(hrng.start,hrng.end):
            hi = self.head_infos[j]
            base_heads.append(cast(hi.var_ptr, VarType))
        head_vars.append(base_heads)
    return head_vars

# -- TODO: These probably shouldn't output a flat array  --

@njit(i8[::1](CREFuncType), cache=True)
def get_head_var_ptrs(self):
    head_var_ptrs = np.empty(self.n_args,dtype=np.int64)
    for i, hrng in enumerate(self.head_ranges):
        for j in range(hrng.start,hrng.end):
            hi = self.head_infos[j]
            head_var_ptrs[i] = hi.var_ptr
    return head_var_ptrs

@overload_attribute(CREFuncTypeClass, "head_var_ptrs")
def overload_head_var_ptrs(self):
    # Not code: returning implementation above
    return get_head_var_ptrs.py_func

# ----------------------------------

@overload_attribute(CREFuncTypeClass, "origin")
def overload_origin(self):
    # Note: This is is dereferencing a weak pointer, but should be safe
    #   since the CREFunc should keep a reference to it's origin
    def impl(self):
        return cast(self.origin_data.ptr, CREFuncType)
    return impl

@overload_attribute(CREFuncTypeClass, "name")
def overload_name(self):
    def impl(self):
        return self.origin_data.name
    return impl

@overload_attribute(CREFuncTypeClass, "expr_template")
def overload_expr_template(self):
    def impl(self):
        return self.origin_data.expr_template
    return impl

@overload_attribute(CREFuncTypeClass, "shorthand_template")
def overload_shorthand_template(self):
    def impl(self):
        return self.origin_data.shorthand_template
    return impl

# ------------------------------
# : CREFunc initialization

@intrinsic
def _get_global_fn_addr(typingctx, literal_name):    
    name = literal_name.literal_value
    def codegen(context, builder, sig, args):
        mod = builder.module
        # Actual type doesn't matter
        fnty = ir.FunctionType(cgutils.voidptr_t, [cgutils.voidptr_t])
        fn = cgutils.get_or_insert_function(mod, fnty, name)
        addr = builder.ptrtoint(fn, cgutils.intp_t)
        return addr
        
    sig = i8(literal_name)
    return sig, codegen

#################Garbage###################3
SIZEOF_NRT_MEMINFO = 48
@njit(types.void(i8),cache=True)
def cf_del_inject(data_ptr):
    cf = cast(data_ptr-SIZEOF_NRT_MEMINFO, CREFuncType)

cf_del_inject_addr = _get_wrapper_address(cf_del_inject, types.void(i8))
ll.add_symbol("CRE_cf_del_inject", cf_del_inject_addr)

##############################################


head_info_size = head_info_type.dtype.itemsize
instr_type_size = instr_type.dtype.itemsize

@njit(#types.void(
    # CREFuncType, 
    # types.optional(head_range_type[::1]),
    # types.optional(head_info_type[::1]),
    # types.optional(instr_type[::1])),
     cache=True)
def rebuild_buffer(self, head_ranges=None, head_infos=None, prereq_instrs=None):
    ''' 'head_ranges', 'head_infos', and 'prereq_instrs' are held in a single
         buffer to ensure data contiguity. This function rebuilds that buffer.
    '''

    n_args = self.n_args if(head_ranges is None) else len(head_ranges)
    n_heads = self.n_args if(head_infos is None) else len(head_infos)
    n_prereqs = 0 if prereq_instrs is None else len(prereq_instrs)

    # Layout: head_ranges | head_infos | prereq_instrs
    L_ = L = n_args + (head_info_size>>2)*n_heads
    L += (instr_type_size>>2) * n_prereqs

    buff = np.zeros(L, dtype=np.uint32)
    if(head_ranges is not None):
        buff[:n_args] = head_ranges[:].view(np.uint32)
    if(head_infos is not None):
        buff[n_args: L_] = head_infos[:].view(np.uint32)
    if(prereq_instrs is not None):
        buff[L_: L] = prereq_instrs[:].view(np.uint32)

    self.head_ranges = buff[:n_args].view(_head_range_type)
    self.head_infos = buff[n_args: L_].view(_head_info_type)
    self.prereq_instrs = buff[L_:L].view(_instr_type)


cached_repr_const_symbols = {}
def ensure_repr_const(typ):
    ''' Implements various jitted repr() functions for various kinds
        of constants that could be part of a CREFunc composition and
        writes their implementations to global variables. These are
        used to str() or repr() CREFuncs.
    '''
    if(isinstance(typ, (Fact,CREObjTypeClass))):
        typ = None
    if(typ not in cached_repr_const_symbols):
        sig = unicode_type(CREObjType, i8)
        if(typ is None):
            # TODO: For now str isn't overloaded for fact types so ignore this
            @njit(sig, cache=True)
            def repr_const(cf, i):
                # Conditional prevents return value from being None
                if(i > 0):
                    raise NotImplementedError("str not implemented for fact types")  
                return ""

        #TODO SHOULD PROPERLY IMPLEMENT REPR For int, float, unicode_type
        elif(typ == unicode_type):
            @njit(sig, cache=True)
            def repr_const(cf, i):
                v = cre_obj_get_item(cf, typ, i8(1+(i<<1)+1))
                return f"'{str(v)}'"
        elif(typ == f8):
            @njit(sig, cache=True)
            def repr_const(cf, i):
                v = cre_obj_get_item(cf, typ, i8(1+(i<<1)+1))

                # Represent round numbers as integers
                if(int(v) == v):
                    return str(int(v))
                else:
                    return str(v)
        else:
            @njit(sig, cache=True)
            def repr_const(cf, i):
                v = cre_obj_get_item(cf, typ, i8(1+(i<<1)+1))
                return str(v)
                
        addr = _get_wrapper_address(repr_const, sig)
        name = f"CREFunc_repr_const_{repr(typ)}"
        ll.add_symbol(name, addr)
        cached_repr_const_symbols[typ] = name
    return cached_repr_const_symbols[typ]

head_info_arr_type = head_info_type[::1]

@generated_jit(nopython=True)
def make_repr_const_addrs(cf, arg_types):
    n_args = len(arg_types)
    # print("<<", arg_types, n_args)
    repr_const_symbols = tuple(ensure_repr_const(at) for  at in arg_types)    
    unroll_inds = tuple(range(n_args))
    if(n_args == 0):
        def impl(cf, arg_types):
            return np.zeros(0, dtype=np.int64)
    else:
        def impl(cf, arg_types):
            repr_const_addrs = np.zeros(n_args, dtype=np.int64)
            for i in literal_unroll(unroll_inds):
                repr_const_addrs[i] = _get_global_fn_addr(repr_const_symbols[i])
            return repr_const_addrs
    return impl


@generated_jit(nopython=True)
def cre_func_ctor(_cf_type, name, expr_template, shorthand_template, is_ptr_func):
    cf_type = _cf_type.instance_type
    fd = cf_type._field_dict
    arg_types = cf_type.arg_types
    n_args = len(arg_types)
    chr_mbr_attrs = ["return_val"]
    for i in range(n_args):
        # NOTE: order of a{i} then h{i} matters.
        chr_mbr_attrs.append(f'a{i}')
        chr_mbr_attrs.append(f'h{i}')
    for i in range(n_args):
        chr_mbr_attrs.append(f'ref{i}')
    chr_mbr_attrs = tuple(chr_mbr_attrs)
    
    # Note: beginning is padded with "" because cannot literal_unroll empty Tuple()
    #  i.e. this allows for the case of a CREFunc with no arguments
    repr_const_symbols = tuple([ensure_repr_const(f8)]+[ensure_repr_const(at) for  at in arg_types])    
    unroll_inds = tuple(range(n_args+1))

    prefix = cf_type.symbol_prefix
    method_names = (
        f"{prefix}_call_heads",
        f"{prefix}_call_self",
        f"{prefix}_resolve_heads",
    )
    # return_t_id = cre_context().get_t_id(cf_type.return_type)

    def impl(_cf_type, name, expr_template, shorthand_template, is_ptr_func):
        cf = new(cf_type)
        cf.idrec = encode_idrec(T_ID_FUNC, 0, 0xFF)
        cf.n_args = n_args
        cf.n_root_args = n_args
        cf.root_arg_infos = np.zeros(n_args, dtype=arg_infos_type)
        cf.is_initialized = False
        cf.is_ptr_func = is_ptr_func
        set_chr_mbrs(cf, chr_mbr_attrs)

        # Point the return_data_ptr to 'return_val'
        ret_t_id,_,return_data_ptr = cre_obj_get_item_t_id_ptr(cf, 0)

        if(ret_t_id == T_ID_STR):
            #Note: Bad things happen if unicode string slots left NULL
            cre_obj_set_item(cf,0,"")

        cf.return_data_ptr = return_data_ptr
        cf.return_t_id = ret_t_id

        self_ptr = cast(cf, i8)

        # Init: head_ranges | head_infos | prereq_instrs
        rebuild_buffer(cast(cf,CREFuncType))

        # Fill head_infos for a newly instantiated CREFunc
        head_infos = cf.head_infos
        for i in range(n_args):
            hi = head_infos[i]
            hi.cf_ptr = self_ptr
            hi.arg_ind = u4(i)
            hi.type = ARGINFO_VAR
            hi.has_deref = 0
            _,_,arg_data_ptr = cre_obj_get_item_t_id_ptr(cf, 1+(i<<1))
            t_id, m_id, head_data_ptr = cre_obj_get_item_t_id_ptr(cf, 1+(i<<1)+1)
            hi.base_t_id = t_id
            hi.head_t_id = t_id
            hi.arg_data_ptr = arg_data_ptr
            hi.head_data_ptr = head_data_ptr

            _,_,elm_aft_ptr = cre_obj_get_item_t_id_ptr(cf, 1+(i<<1)+2)
            hi.head_size = u4(elm_aft_ptr-head_data_ptr)

            if(t_id == T_ID_STR):
                hi.ref_kind = REFKIND_UNICODE
                #Note: Bad things happen if unicode string slots left NULL
                cre_obj_set_item(cf,1+(i<<1)+1,"")
            elif(m_id != PRIMITIVE_MBR_ID):
                hi.ref_kind = REFKIND_STRUCTREF
            else:
                hi.ref_kind = REFKIND_PRIMATIVE

            cf.head_ranges[i].start = i
            cf.head_ranges[i].end = i+1

        # Load addresses of methods from globals 
        # if(not is_ptr_func):
        cf.call_heads_addr = _get_global_fn_addr(method_names[0])
        cf.call_self_addr = _get_global_fn_addr(method_names[1])
        cf.resolve_heads_addr = _get_global_fn_addr(method_names[2])        

        repr_const_addrs = np.zeros(n_args, dtype=np.int64)
        for i in literal_unroll(unroll_inds):
            # Note unroll inds loops one past N_
            if(i != 0):
                repr_const_addrs[i-1] = _get_global_fn_addr(repr_const_symbols[i])
        # return repr_const_addrs
        # repr_const_addrs = make_repr_const_addrs(cf, arg_types)
        cf.origin_data = OriginData(cast(cf,i8), name, expr_template, shorthand_template, repr_const_addrs)        
        cf.is_composed = False
        cf.depth = 1

        casted = cast(cf, CREFuncType)
        return casted 
    return impl

# -------------------------------------
# : Helper Functions 

@njit(CREFuncType(i8), cache=True)
def func_from_ptr(ptr):
    return cast(ptr, CREFuncType)

@njit(i8(CREFuncType), cache=True)
def func_get_ptr(self):
    return cast(self, i8)


# --------------------------------------------------------------
# : Copy

from cre.obj import copy_cre_obj
@njit(cache=False)
def cre_func_copy(cf):
    # Make a copy of the CreFunc via a memcpy (plus incref obj members)
    cpy = copy_cre_obj(cf)

    # Find the the byte offset between the CREFunc and its copy
    cf_ptr = cast(cf, i8)
    cpy_ptr = cast(cpy, i8)
    cpy_delta = cpy_ptr-cf_ptr

    # Make a copy of base_to_head_infos
    # base_to_head_infos = List.empty_list(head_info_arr_type)
    old_head_ranges = cf.head_ranges.copy()
    old_head_infos = cf.head_infos.copy()
    old_prereq_instrs = cf.prereq_instrs.copy()

    # Nullify these attributes since we don't want the pointers from the 
    #  original to get decref'ed on assignment
    _nullify_attr(cpy, 'head_ranges')
    _nullify_attr(cpy, 'head_infos')
    _nullify_attr(cpy, 'root_arg_infos')
    _nullify_attr(cpy, 'origin_data')
    _nullify_attr(cpy, 'prereq_instrs')

    # Rebuild head_ranges, head_infos, prereq_instrs from common buffer
    cpy_generic = cast(cpy, CREFuncType)
    rebuild_buffer(cpy_generic, old_head_ranges, old_head_infos, old_prereq_instrs)

    # Make the arg_data_ptr and head_data_ptr point to the copy
    for i, head_info in enumerate(cf.head_infos):
        cpy.head_infos[i] = head_info
        if(head_info.cf_ptr == cf_ptr):
            cpy.head_infos[i].cf_ptr = cpy_ptr
            cpy.head_infos[i].arg_data_ptr = head_info.arg_data_ptr+cpy_delta
            cpy.head_infos[i].head_data_ptr = head_info.head_data_ptr+cpy_delta

    cpy.return_data_ptr = cf.return_data_ptr+cpy_delta
    cpy.root_arg_infos = cf.root_arg_infos.copy()
    cpy.origin_data = cf.origin_data
    cpy.is_initialized = False

    # print("::", f"{cf_ptr}:{cast(cf.root_arg_infos, i8)}","|", f"{cpy_ptr}:{cast(cpy.root_arg_infos, i8)}")
    # print("&&>",cf_ptr,cf.return_data_ptr,cf.return_data_ptr-cf_ptr)
    # print("&&<",cpy_ptr,cpy.return_data_ptr,cpy.return_data_ptr-cpy_ptr)

    return cpy

@njit(CREFuncType(CREFuncType), cache=True)
def cre_func_copy_generic(cf):
    return cre_func_copy(cast(cf, CREFuncType))


@njit(CREFuncType(CREFuncType),cache=True)
def cre_func_deep_copy_generic(cf):
    '''Makes a deep copy of a CREFunc Object.'''

    if(len(cf.prereq_instrs)==0):
        return cre_func_copy_generic(cf)

    # Initialize Data Structures
    stack = List()
    i = 0 
    func_copies = List.empty_list(CREFuncType)
    remap = Dict.empty(i8,i8)

    keep_looping = True
    while(keep_looping):
        arg_info = cf.root_arg_infos[i]
        if(arg_info.type == ARGINFO_FUNC):
            # Push Stack i.e. Recurse Arg
            stack.append((cf, i+1, func_copies))
            cf = cast(arg_info.ptr, CREFuncType)
            func_copies = List.empty_list(CREFuncType)
            i = 0 
        else:
            # Skip Arg
            i += 1

        # When past end arg of deepest func, copy it. Then pop prev frame from stack.
        while(i >= len(cf.root_arg_infos)):
            cpy = cre_func_copy_generic(cf)

            # Set the "ref{i}" and root_arg_info.ptr to copies of func children 
            j = 0 
            for k, inf in enumerate(cpy.root_arg_infos):
                if(inf.type == ARGINFO_FUNC):
                    sub_op = func_copies[j]
                    inf.ptr = cast(sub_op, i8)
                    cre_obj_set_item(cpy, i8(1+cpy.n_root_args*2 + k), sub_op)
                    j += 1

            remap[cast(cf, i8)] = cast(cpy, i8)
            remap[cf.return_data_ptr] = cpy.return_data_ptr

            # Remap pointers in head_infos 
            for hi, hi_cpy in zip(cf.head_infos, cpy.head_infos):
                hi_cpy.cf_ptr = remap.get(hi.cf_ptr, hi_cpy.cf_ptr)
                hi_cpy.arg_data_ptr = remap.get(hi.arg_data_ptr, hi_cpy.arg_data_ptr)
                hi_cpy.head_data_ptr = remap.get(hi.head_data_ptr, hi_cpy.head_data_ptr)
                remap[hi.arg_data_ptr] = hi_cpy.arg_data_ptr
                remap[hi.head_data_ptr] = hi_cpy.head_data_ptr

            # Remap pointers in prereq_instrs
            for instr, instr_cpy in zip(cf.prereq_instrs, cpy.prereq_instrs):
                instr_cpy.cf_ptr = remap.get(instr.cf_ptr, instr_cpy.cf_ptr)
                instr_cpy.return_data_ptr = remap.get(instr.return_data_ptr, instr_cpy.return_data_ptr)

            # End Case: Stack exhausted
            if(len(stack) == 0):
                keep_looping = False
                break

            # Pop off stack i.e. equivalent to returning recursive call
            cf, i, func_copies = stack.pop(-1)
            func_copies.append(cpy)
    return cpy

cf_deep_copy_ep = cre_func_deep_copy_generic.overloads[(CREFuncType,)].entry_point

#--------------------------------------------------------------------
# Construction Functions


from numba.core.typing.typeof import typeof
set_const_arg_overloads = {}
def set_const_arg_impl(_val, use_ep=False):
    nb_val_type = typeof(_val) if not isinstance(_val, types.Type) else _val
    if(nb_val_type not in set_const_arg_overloads):
        val_t_id = cre_context().get_t_id(nb_val_type)
        sig = types.void(CREFuncType, i8, nb_val_type)
        @njit(sig,cache=True)
        def _set_const_arg(self, i, val):
            self.is_initialized = False

            head_infos = self.head_infos
            start = self.head_ranges[i].start
            end = self.head_ranges[i].end
            for j in range(start,end):
                cf = cast(head_infos[j].cf_ptr, CREFuncType)
                arg_ind = head_infos[j].arg_ind

                # NOTE: Don't need to bother checking here because if it will fail it 
                #  will fail at compile time
                # if(head_infos[j].base_t_id != val_t_id and not self.is_ptr_func):
                #     raise TypeError("Constant's type doesn't match composing CREFunc's argument type.")

                head_infos[j].type = ARGINFO_CONST
                head_infos[j].has_deref = 0
                # set 'h{i}' to val
                cre_obj_set_item(cf, i8(1+(arg_ind<<1)+1), val)
                # set 'ref{i}' to None
                cre_obj_set_item(cf, i8(1+cf.n_root_args*2 + arg_ind), None)

                cf.root_arg_infos[arg_ind].type = ARGINFO_CONST
                cf.root_arg_infos[arg_ind].has_deref = 0
                cf.root_arg_infos[arg_ind].ptr = head_infos[j].head_data_ptr
                cf.root_arg_infos[arg_ind].t_id = head_infos[j].head_t_id

        entry_point = next(iter(_set_const_arg.overloads.values())).entry_point
        set_const_arg_overloads[nb_val_type] = (_set_const_arg, entry_point)
    if(use_ep):
        return set_const_arg_overloads[nb_val_type][1]
    else:
        return set_const_arg_overloads[nb_val_type][0]


get_const_root_arg_overloads = {}
def get_const_root_arg_impl(_typ_or_t_id, use_ep=False):
    _typ = _typ_or_t_id
    if(isinstance(_typ_or_t_id, int)):
        context = cre_context()
        _typ = context.get_type(t_id=_typ_or_t_id)

    sig = _typ(CREFuncType, i8)
    @njit(sig,cache=True)
    def _get_const_root_arg(self, i):
        # print(self.root_arg_infos)
        v = cre_obj_get_item(self, _typ, i8(1+(i<<1)+1))
        return v
        # return _load_ptr(_typ, self.root_arg_infos[i].ptr)
    entry_point = next(iter(_get_const_root_arg.overloads.values())).entry_point
    get_const_root_arg_overloads[_typ] = (_get_const_root_arg, entry_point)
    if(use_ep):
        return get_const_root_arg_overloads[_typ][1]
    else:
        return get_const_root_arg_overloads[_typ][0]


        



@generated_jit(cache=True)
def set_const_arg(self, i, val):
    impl = set_const_arg_impl(val)
    return impl

@njit(types.void(CREFuncType,i8,VarType), cache=True)
def set_var_arg(self, i, var):
    self.is_initialized = False
    head_infos = self.head_infos
    start = self.head_ranges[i].start
    end = self.head_ranges[i].end
    
    hd = u1(len(var.deref_infos) > 0)
    for j in range(start,end):
        cf = cast(head_infos[j].cf_ptr, CREFuncType)
        arg_ind = head_infos[j].arg_ind

        if(head_infos[j].base_t_id != var.head_t_id and not self.is_ptr_func):
            raise TypeError("Var's head_type doesn't match composing CREFunc's argument type.")

        if(head_infos[j].has_deref):
            old_head_var = cast(head_infos[j].var_ptr, VarType)
            head_var = var_extend(var, old_head_var.deref_infos)
            head_infos[j].has_deref = u1(1)
        else:
            head_var = var
            head_infos[j].has_deref = u1(hd)

        head_infos[j].var_ptr = cast(head_var, i8)
        head_infos[j].type = ARGINFO_VAR

        # set 'ref{i}' to var
        cre_obj_set_item(cf, i8(1+cf.n_root_args*2 + arg_ind), head_var)

        cf.root_arg_infos[arg_ind].type = ARGINFO_VAR
        cf.root_arg_infos[arg_ind].has_deref = head_infos[j].has_deref
        cf.root_arg_infos[arg_ind].ptr = cast(head_var, i8)
        cf.root_arg_infos[arg_ind].t_id = head_infos[j].head_t_id

set_var_arg_ep = set_var_arg.overloads[(CREFuncType,i8,VarType)].entry_point


@njit(types.void(CREFuncType,i8,CREFuncType), cache=True, debug=False)
def set_func_arg(self, i, func):
    self.is_initialized = False
    head_infos = self.head_infos
    start = self.head_ranges[i].start
    end = self.head_ranges[i].end

    cf_ptr = cast(func, i8)
    for j in range(start,end):
        cf = cast(head_infos[j].cf_ptr, CREFuncType)
        arg_ind = head_infos[j].arg_ind

        if(head_infos[j].base_t_id != func.return_t_id and not self.is_ptr_func):
            raise TypeError("Argument CREFunc's head_type doesn't match composing CREFunc's argument type.")

        head_infos[j].cf_ptr = cf_ptr
        head_infos[j].has_deref = 0
        head_infos[j].type = ARGINFO_FUNC_UNEXPANDED

        # Set ref{i} to func
        cre_obj_set_item(cf, i8(1+cf.n_root_args*2 + arg_ind), func)

        cf.root_arg_infos[arg_ind].type = ARGINFO_FUNC
        cf.root_arg_infos[arg_ind].has_deref = 0
        cf.root_arg_infos[arg_ind].ptr = cf_ptr
        cf.root_arg_infos[arg_ind].t_id = head_infos[j].head_t_id

set_func_arg_ep = set_func_arg.overloads[(CREFuncType,i8,CREFuncType)].entry_point


cf_ind_tup_t = Tuple((CREFuncType,i8))
@njit(instr_type[::1](CREFuncType),cache=True)
def build_instr_set(self):
    ''' Rebuilds  prereq_intrs for a CREFunc '''
    # NOTE: Since we can't cache recursive functions use a stack to 
    #  do a DF traversal of funcs and build the instr_set which defines
    #  the execution order of this func's child funcs.

    instrs = List.empty_list(instr_type)
    stack = List.empty_list(cf_ind_tup_t)
    cf = self
    i = 0 
    n_iter = 0
    keep_looping = True
    while(keep_looping):
        arg_info = cf.root_arg_infos[i]
        assert arg_info.ptr != cast(cf, i8), "CREFunc has self reference"
        if(arg_info.type == ARGINFO_FUNC):
            t_id, m_id, head_data_ptr = cre_obj_get_item_t_id_ptr(cf,(1+(i<<1)+1))
            instr = np.zeros(1,dtype=instr_type)[0]
            instr.cf_ptr = arg_info.ptr
            instr.return_data_ptr = head_data_ptr

            if(t_id == T_ID_STR):
                instr.size = _sizeof_type(unicode_type) #Hard-coded ... see commented out part below
                instr.ref_kind = REFKIND_UNICODE
            elif(m_id != PRIMITIVE_MBR_ID):
                instr.size = _sizeof_type(CREFuncType)
                instr.ref_kind = REFKIND_STRUCTREF
            else:
                instr.size = _sizeof_type(i8)
                instr.ref_kind = REFKIND_PRIMATIVE

            # Set the size to be the difference between the data_ptrs for 
            #  the return value and the a0 member 
            # _, _, ret_data_ptr = cre_obj_get_item_t_id_ptr(cf,0)
            # _, _, first_data_ptr = cre_obj_get_item_t_id_ptr(cf,1)
            # instr.size = u4(first_data_ptr-ret_data_ptr)

            instrs.append(instr)
            stack.append((cf, i+1))
            cf = cast(arg_info.ptr, CREFuncType)
            i = 0 
        else:
            i += 1

        while(i >= len(cf.root_arg_infos)):
            if(len(stack) == 0):
                keep_looping = False
                break
            cf, i = stack.pop(-1)

        n_iter += 1
        if(n_iter >= 128):
            raise RuntimeError("CREFunc max construction depth exceeded.")

    prereq_instrs = np.zeros(len(instrs), dtype=instr_type)
    j = 0
    for i in range(len(instrs)-1,-1,-1): #enumerate(reversed(instrs)):
        if(instrs[i].size > 100): raise RuntimeError("Bad instr")

        prereq_instrs[j] = instrs[i]
        j += 1
    return prereq_instrs

i8_arr = i8[::1]
head_info_lst = ListType(head_info_type)
@njit(types.void(CREFuncType),cache=True)
def reinitialize(self):
    ''' Reinitializes a CREFunc after a composition has been made with 
        a sequence of calls to set_[func,var,const]_arg(). This function 
        ensures that head_ranges, head_infos, and prereq_instrs are all
        properly set to reflect the composition. 
    '''
    if(self.is_initialized): return

    max_arg_depth = 0

    base_var_map = Dict.empty(i8, head_info_lst)
    # Go through the current HEAD_INFOS looking for entries with 
    #  kind ARGINFO_VAR or ARGINFO_FUNC_UNEXPANDED 
    for hrng in self.head_ranges:
        for j in range(hrng.start,hrng.end):
            head_info = self.head_infos[j]

            # For ARGINFO_VAR kinds insert base_ptr into the base_var_map
            if(head_info.type == ARGINFO_VAR):
                var = cast(head_info.var_ptr, VarType)
                base_ptr = var.base_ptr
                if(base_ptr not in base_var_map):
                    base_var_map[base_ptr] = List.empty_list(head_info_type)
                base_var_map[base_ptr].append(head_info)

            # For ARGINFO_FUNC_UNEXPANDED kind insert the base_ptrs of all of the
            #  CRE_Funcs's base vars into base_var_map
            elif(head_info.type == ARGINFO_FUNC_UNEXPANDED):
                cf = cast(head_info.cf_ptr, CREFuncType)
                for hrng_k in cf.head_ranges:
                    for n  in range(hrng_k.start, hrng_k.end):
                        head_info_n = cf.head_infos[n]
                        var = cast(head_info_n.var_ptr, VarType)
                        base_ptr = var.base_ptr
                        if(base_ptr not in base_var_map):
                            base_var_map[base_ptr] = List.empty_list(head_info_type)
                        base_var_map[base_ptr].append(head_info_n)
                max_arg_depth = max(cf.depth, max_arg_depth)

    # Make new head_ranges according to base_var_map
    n_bases = len(base_var_map)
    head_ranges = np.zeros(n_bases, dtype=head_range_type)
    n_heads = 0
    for i, base_head_infos in enumerate(base_var_map.values()):
        head_ranges[i].start = n_heads
        head_ranges[i].end = n_heads+len(base_head_infos)
        n_heads += len(base_head_infos)

    # Make new head_infos according to base_var_map
    head_infos = np.zeros(n_heads,dtype=head_info_type)
    c = 0
    for i, (base_ptr, base_head_infos) in enumerate(base_var_map.items()):
        base_var = cast(base_ptr, VarType)
        for j in range(len(base_head_infos)):
            head_infos[c] = base_head_infos[j]
            head_infos[c].base_t_id = base_var.base_t_id
            c += 1

    self.n_args = n_bases

    prereq_instrs = build_instr_set(self)
    rebuild_buffer(self, head_ranges, head_infos, prereq_instrs)

    # Record if the CREFunc was modified enough from it's base definition
    #  that it should be considered 'is_composed'. Also check 'has_any_derefs'.
    any_hd = False
    is_composed = False
    for inf in self.root_arg_infos:
        any_hd = any_hd | inf.has_deref
        if(any_hd or inf.type == ARGINFO_FUNC or inf.type == ARGINFO_CONST):
            is_composed = True
    self.has_any_derefs = any_hd
    self.is_composed = is_composed
    self.depth = self.depth + max_arg_depth
  

#--------------------------------------------------------------------
# Execution Functions

#--------------------------------
# Set Arguments
        
from numba.core.typing.typeof import typeof
set_base_arg_val_overloads = {}
def set_base_arg_val_impl(_val):
    # Get the type of _val or if a typeref was given use that
    nb_val_type = typeof(_val) if not isinstance(_val, types.Type) else _val

    # If is a Fact or other object type upcast to CREObjType
    nb_val_type = CREObjType if isinstance(nb_val_type, CREObjTypeClass) else nb_val_type
    # Compile the implementation if it doesn't exist
    if(nb_val_type not in set_base_arg_val_overloads):        
        sig = types.void(CREFuncType, u8, nb_val_type)
        if(nb_val_type is CREObjType):
            # Fact case
            @njit(sig,cache=True)
            def _set_base_arg_val(self, i, val):
                start = self.head_ranges[i].start
                end = self.head_ranges[i].end
                for j in range(start,end):
                    if(self.head_infos[j].has_deref):
                        _store_safe(nb_val_type, self.head_infos[j].arg_data_ptr, val)
                    else:
                        _store_safe(nb_val_type, self.head_infos[j].head_data_ptr, val)
        else:
            # Primitive case
            @njit(sig,cache=True)
            def _set_base_arg_val(self, i, val):
                start = self.head_ranges[i].start
                end = self.head_ranges[i].end
                for j in range(start,end):
                    _store_safe(nb_val_type, self.head_infos[j].head_data_ptr, val)
        set_base_arg_val_overloads[nb_val_type] = _set_base_arg_val
    return set_base_arg_val_overloads[nb_val_type]

set_obj_base_val = set_base_arg_val_impl

@njit(types.void(CREFuncType, u4, i8[::1]), cache=True, locals={"i" : u4,"k":u4, "j": u4})
def set_heads_from_data_ptrs(cf, i, data_ptrs):
    for k, j in enumerate(range(cf.head_ranges[i].start, cf.head_ranges[i].end)):
        hi = cf.head_infos[j]
        if(hi.ref_kind==REFKIND_UNICODE): 
            # _decref_structref(_load_ptr(unicode_type, hi.head_data_ptr))
            _incref_structref(_load_ptr(unicode_type, data_ptrs[k]))
        elif(hi.ref_kind==REFKIND_STRUCTREF):
            # _decref_ptr(_load_ptr(i8, hi.head_data_ptr))
            _incref_ptr(_load_ptr(i8, data_ptrs[k]))

        _memcpy(hi.head_data_ptr, data_ptrs[k], hi.head_size)

#--------------------------------
# Call

call_self_f_type = types.FunctionType(u1(CREFuncType))

@njit(u1(CREFuncType), locals={"i":u8}, cache=True)
def cre_func_call_self(self):
    ''' Calls the CREFunc including any dependant CREFuncs in its composition
        Each CREFunc's a{i} and h{i} characteristic members are used as
        intermediate memory spaces for the results of these calculations.
    '''
    for i in range(len(self.prereq_instrs)):
        instr = self.prereq_instrs[i]
        cf = cast(instr.cf_ptr, CREFuncType)

        status = _func_from_address(call_self_f_type, cf.call_self_addr)(cf)
        if(status > 1): return status

        if(instr.ref_kind==REFKIND_UNICODE):
            new_obj = _load_ptr(unicode_type, cf.return_data_ptr)   
            _incref_structref(new_obj)
        elif(instr.ref_kind==REFKIND_STRUCTREF):
            new_obj_ptr = _load_ptr(i8, cf.return_data_ptr)   
            _incref_ptr(new_obj_ptr)

        _memcpy(instr.return_data_ptr, cf.return_data_ptr, instr.size)
    status = _func_from_address(call_self_f_type, self.call_self_addr)(self)
    return status

CREFunc_method(CREFuncType, u1(CREFuncType), "call_self")(cre_func_call_self)


@njit(u1(CREFuncType), locals={"i":u8}, cache=True)
def cre_func_resolve_call_self(self):
    ''' As 'cre_func_call_self' but also dereferences in the composition's
        head variables. For instance like my_cf(a.nxt.B, b.nxt.A).
    '''
    for i in range(len(self.prereq_instrs)):
        
        instr = self.prereq_instrs[i]
        cf = cast(instr.cf_ptr, CREFuncType)
        if(cf.has_any_derefs):
            status = _func_from_address(call_self_f_type, cf.resolve_heads_addr)(cf)
            if(status): return status
        status = _func_from_address(call_self_f_type, cf.call_self_addr)(cf)
        if(status > 1): return status

        if(instr.ref_kind==REFKIND_UNICODE):
            new_obj = _load_ptr(unicode_type, cf.return_data_ptr)   
            _incref_structref(new_obj)
        elif(instr.ref_kind==REFKIND_STRUCTREF):
            new_obj_ptr = _load_ptr(i8, cf.return_data_ptr)   
            _incref_ptr(new_obj_ptr)
    
        _memcpy(instr.return_data_ptr, cf.return_data_ptr, instr.size)

    if(self.has_any_derefs):
        status = _func_from_address(call_self_f_type, self.resolve_heads_addr)(self)
        if(status): return status

    status = _func_from_address(call_self_f_type, self.call_self_addr)(self)
    if(status > 1): return status
    return status

CREFunc_method(CREFuncType, u1(CREFuncType), "resolve_call_self")(cre_func_resolve_call_self)


# TODO: Make this
# def identity_call_self():
#     if(self.has_any_derefs):
#         status = _func_from_address(call_self_f_type, self.resolve_heads_addr)(self)
#         if(status): return status

#     if(instr.ref_kind==REFKIND_UNICODE):
#         new_obj = _load_ptr(unicode_type, cf.return_data_ptr)   
#         _incref_structref(new_obj)
#     elif(instr.ref_kind==REFKIND_STRUCTREF):
#         new_obj_ptr = _load_ptr(i8, cf.return_data_ptr)   
#         _incref_ptr(new_obj_ptr)

#     _memcpy(cf.return_data_ptr, cf.head_infos[0].head_data_ptr return_data_ptr, instr.size)
#     _store_safe(return_type, self.return_data_ptr, return_val)



cs_gbl_name = 'CREFuncType_call_self'
res_cs_gbl_name = 'CREFuncType_resolve_call_self'

@njit
def get_best_call_self(self, ignore_derefs=False):
    '''Returns the fastest implementation of call_self for a CREFunc.
        In the best case, if the CREFunc doesn't have other CREFuncs in it's 
        composition and no head Vars with dereference instructions then the 
        fastest implementation is it's type specific implementation.
    '''
    if(not ignore_derefs and self.has_any_derefs):
        # Use cre_func_resolve_call_self
        cs_addr = _get_global_fn_addr(res_cs_gbl_name)
    elif(len(self.prereq_instrs) > 0):
        # Use cre_func_call_self
        cs_addr = _get_global_fn_addr(cs_gbl_name)
    else:
        # Use base type-specific implementation
        cs_addr = self.call_self_addr
    return _func_from_address(call_self_f_type, cs_addr)

#--------------------------------
# Get Return Value
    
from numba.core.typing.typeof import typeof
get_return_val_overloads = {}
# i8_x6 = UniTuple(i8,6)
def get_return_val_impl(val_type):
    ''' Implementation for extracting return value of 'val_type' from CREFunc '''
    nb_val_type = typeof(val_type) if not isinstance(val_type, types.Type) else val_type
    if(nb_val_type not in get_return_val_overloads):
        @njit(nb_val_type(CREFuncType),cache=True, inline='always')
        def _get_return_val(self):
            val = _load_ptr(nb_val_type, self.return_data_ptr)
            # _incref_structref(val)
            # _store(i8_x6, self.return_data_ptr, (0,0,0,0,0,0))
            # print(_load_ptr(nb_val_type, self.return_data_ptr))
            return val
        get_return_val_overloads[nb_val_type] = _get_return_val
    return get_return_val_overloads[nb_val_type]


#--------------------------------------------------------------------
@generated_jit
def _set_var_helper(kind, cf, i, arg):
    '''Helper function for applying set_[var,func,const]_arg depending on 'kind' '''
    # SentryLiteralArgs(['kind']).for_function(_set_var_helper).bind(kind, cf, i, arg)
    kind = kind.literal_value
    if(kind==ARGINFO_VAR):
        def impl(kind, cf, i, arg):    
            set_var_arg(cf, i, cast(arg, VarType))  
    elif(kind==ARGINFO_FUNC):
        def impl(kind, cf, i, arg):    
            set_func_arg(cf, i, cast(arg, CREFuncType))
    elif(kind==ARGINFO_CONST):
        def impl(kind, cf, i, arg):    
            set_const_arg(cf, i, arg)
    return impl
    
# Overload call/compose
from numba.types import unliteral
def overload_compose_codegen(cf_type, arg_types):
    ''' Codegen for composing a CREFunc from jitted code'''
    kinds = []
    for i, at in enumerate(arg_types):
        if(isinstance(at, VarTypeClass)):
            kinds.append(int(ARGINFO_VAR))
        elif(isinstance(at,CREFuncTypeClass)):
            kinds.append(int(ARGINFO_FUNC))
        else:
            kinds.append(int(ARGINFO_CONST))

    most_precise_cf_type = CREFuncTypeClass(return_type=cf_type.return_type, name="FuncComp")
    inds = tuple([*range(len(arg_types))])
    kinds = tuple(kinds)
    def impl(cf, *args):
        _cf = cast(cre_func_deep_copy_generic(cf), most_precise_cf_type)#cre_func_copy(cf)
        for i in literal_unroll(inds):
            _set_var_helper(kinds[i],_cf,i,args[i])
        reinitialize(_cf)
        return cast(_cf, most_precise_cf_type)
    return impl


def overload_call_codegen(cf_type, arg_types):
    ''' Codegen for calling a CREFunc from jitted code'''
    if(cf_type.is_composed):
        if(len(arg_types) > 0 and isinstance(arg_types[0],types.BaseTuple)):
            arg_types = tuple(x for x in arg_types[0])

        set_base_impls = tuple(set_base_arg_val_impl(a) for a in arg_types)
        set_base = set_base_impls[0]
        ret_impl = get_return_val_impl(cf_type.return_type)
        range_inds = tuple(u8(i) for i in range(len(arg_types)))
        def impl(cf,*args):
            for i in literal_unroll(range_inds):
                set_base(cf, i, args[i])
            status = cre_func_resolve_call_self(cf)
            if(status == CFSTATUS_NULL_DEREF):
                raise ValueError("CREFunc failed to execute because of a dereferencing error.")
            elif(status == CFSTATUS_ERROR):
                raise ValueError("CREFunc failed to execute because of an internal error.")
            return ret_impl(cf)

    elif(hasattr(cf_type,'dispatchers') and 'call' in cf_type.dispatchers):
        call = cf_type.dispatchers['call']
        def impl(cf, *args):
            return call(*args)
    else:
        fn_type = types.FunctionType(cf_type.return_type(*cf_type.arg_types))
        def impl(cf, *args):
            f = _func_from_address(fn_type, cf.call_heads_addr)
            return f(*args)
    return impl


#-----------------------------------------------------------------
# : CREFunc Overload __call__

@overload_method(CREFuncTypeClass, '__call__')
def overload_call(cf, *args):
    ''' Allows for CREFuncs to be called and composed from jitted code.'''
    cf_type = cf
    if(len(args) > 0 and isinstance(args[0], types.BaseTuple)):
        args = tuple(*args)

    all_const = all_args_are_const(args)
    if(all_const):
        if(not getattr(cf_type, 'return_type',None)):
            raise ValueError("Cannot call CREFunc without return_type")

        return overload_call_codegen(cf, args)
    else:
        if(not getattr(cf_type, 'return_type',None)):
            raise ValueError("Cannot compose CREFunc without return_type")

        return overload_compose_codegen(cf, args)
            
    return impl


#-----------------------------------------------------------------
# : cre_func_str()

rc_fn_typ = types.FunctionType(unicode_type(CREObjType, i8))


@njit(unicode_type(CREFuncType, types.boolean, types.boolean,
            unicode_type), cache=True)
def cre_func_str(self, use_shorthand, use_derefs, ignore_pattern):
    ''' Generates a string representation for a CREFunc.'''

    # print("ignore_cfs", ignore_cfs)

    stack = List()
    cf = self
    i = 0 
    arg_strs = List.empty_list(unicode_type)
    s = ""
    keep_looping = True
    while(keep_looping):
        if(i < len(cf.root_arg_infos)):
            arg_info = cf.root_arg_infos[i]
            if(arg_info.type == ARGINFO_FUNC):
                stack.append((cf, i+1,arg_strs))
                cf = cast(arg_info.ptr, CREFuncType)
                arg_strs = List.empty_list(unicode_type)
                i = 0 
            else:
                if(arg_info.type == ARGINFO_VAR):
                    var = cast(arg_info.ptr, VarType)
                    if(use_derefs):
                        arg_strs.append(str(var)) 
                    else:
                        arg_strs.append(var.alias) 
                elif(arg_info.type == ARGINFO_CONST):
                    addr = cf.origin_data.repr_const_addrs[i]
                    fn = _func_from_address(rc_fn_typ, addr)
                    arg_strs.append(fn(cf,i))
                else:
                    raise ValueError("Bad arginfo type.")
                i += 1
        while(i >= len(cf.root_arg_infos)):
            nd = cf.origin_data

            should_ignore = f"{nd.name}," in ignore_pattern
            if(should_ignore):
                if(cf.n_root_args > 1): 
                    raise ValueError("Cannot ignore printing CREFunc with more than one argument.")
                tmp = "{0}"
            else:
                tmp = nd.shorthand_template if use_shorthand else nd.expr_template
            s = tmp.format(arg_strs)

            if(len(stack) == 0):
                keep_looping = False
                break

            if(use_shorthand and not should_ignore):
                parent_nd = stack[-1][0].origin_data
                parent_tmp = parent_nd.shorthand_template if use_shorthand else parent_nd.expr_template
                if(tmp[-1] != ")" and parent_tmp[-1] != ")"):
                    s = f"({s})"

            cf, i, arg_strs = stack.pop(-1)
            arg_strs.append(s)
    return s

@overload_method(CREFuncTypeClass, '__str__')
@overload(str)
def overload_cre_func_str(self):
    if(self is not CREFuncType): return
    def impl(self):
        return cre_func_str(self, True, True, "")
    return impl

#------------------------------------------------------------------
# : define_CREFunc()

def _standardize_commutes(members):
    '''Standardize 'commutes' and build 'right_commutes' from 'commutes' '''

    # If commutes wasn't given then just assign default empty values (i.e. nothing commutes)
    commutes = members['commutes'] = members.get('commutes',[])
    right_commutes = members['right_commutes'] = members.get('right_commutes',{})
    if('commutes' not in members): return

    # If commutes is True then all args of the same type commute
    arg_types = members['arg_types']
    if(isinstance(commutes,bool)):
        if(commutes == True):
            commutes, d = [], {}
            for i, typ in enumerate(arg_types): 
                d[typ] = d.get(typ,[]) + [i] 
            for typ, inds in d.items():
                commutes += [inds]
        else:
            commutes = []
    else:
        assert(isinstance(commutes,Iterable))

    # Fill in right commutes 
    right_commutes = {}
    for i in range(len(commutes)):
        commuting_set = commutes[i]
        for j in range(len(commuting_set)-1,0,-1):
            right_commutes[commuting_set[j]] = np.array(commuting_set[0:j],dtype=np.int64)
            for k in commuting_set[0:j]:
                typ1, typ2 = arg_types[k], arg_types[commuting_set[j]]
                assert typ1==typ2, \
        f"CREFunc {members['func_name']!r} has invalid 'commutes' member {commutes}. Signature arguments {k} and {commuting_set[j]} have different types {typ1} and {typ2}."
    right_commutes = right_commutes

    members['commutes'] = commutes
    members['right_commutes'] = right_commutes

def _standardize_nopython(members):
    ''' 'nopython_call' and 'nopython_check' default to value of 'no_python' or None '''
    default = members.get('nopython',None)
    members['nopython_call'] = members.get('nopython', default)
    members['nopython_check'] = members.get('nopython', default)

    return members['nopython_call'], members['nopython_check']

def _py_func_unq(py_func):
    codebytes = py_func.__code__.co_code
    return codebytes
    # TODO: ALSO NEED __closure__?
    # print("cb:", codebytes)
    # if py_func.__closure__ is not None:
    #     cvars = tuple([x.cell_contents for x in self._py_func.__closure__])
    #     # Note: cloudpickle serializes a function differently depending
    #     #       on how the process is launched; e.g. multiprocessing.Process
    #     cvarbytes = dumps(cvars)
    # else:
    #     cvarbytes = b''


def _standardize_method(members, method_name):
    ''' Extract the py_funcs associated with 'call' or 'check' and pickle them.'''
    if(method_name in members):
        # Extract and store thes python_func
        func = members.get(method_name)
        py_func = func.py_func if isinstance(func, Dispatcher) else func
        # members[f'{method_name}_pyfunc'] = py_func

        # Speed up by caching cloudpickle.dumps(py_func) inside py_func object
        if(hasattr(py_func,'_cre_cloudpickle_bytes')):
            members[f'{method_name}_bytes'] = py_func._cre_cloudpickle_bytes       
        else:
            # If pickle bytes weren't cached then pickle it 
            cp_bytes = cloudpickle.dumps(py_func)
            members[f'{method_name}_bytes'] = cp_bytes
            py_func._cre_cloudpickle_bytes = cp_bytes

        return members[f'{method_name}_bytes'], _py_func_unq(py_func)
    else:
        return None, None
    
def define_CREFunc(name, members):
    ''' Defines a new CREFuncType '''

    # Ensure that 'call' is present and that 'call' and 'check' are functions
    has_check = 'check' in members
    assert 'call' in members, "CREFunc must have call() defined"
    assert hasattr(members['call'], '__call__'), "call() must be a function"
    assert not (has_check and not hasattr(members['check'], '__call__')), "check() must be a function"

    members['func_name'] = name
    # inp_members = {**members}

    no_raise = members.get('no_raise',False)
    ptr_args = members.get('ptr_args',False)

    # Unpack signature
    signature = members['signature']
    return_type = members['return_type'] = signature.return_type
    if(not ptr_args):
        arg_types = members['arg_types'] = signature.args
    else:
        arg_types = members['arg_types'] = tuple([CREObjType]*len(signature.args))

    members['signature'] = return_type(*arg_types)

    # Standardize commutes and nopython
    _standardize_commutes(members)
    nopy_call, nopy_check = _standardize_nopython(members)

    # Get pickle bytes for 'call' and 'check'
    call_bytes, call_unq =_standardize_method(members, 'call')
    check_bytes, check_unq = _standardize_method(members, 'check')

    # Regenerate the source for this type if wasn't cached or if 'cache=False' 
    unq_args = [name, return_type, arg_types, call_bytes, check_bytes, no_raise, ptr_args]
    gen_src_args = [name, return_type, arg_types, nopy_call, nopy_check, call_bytes, check_bytes, no_raise, ptr_args]
    long_hash = unique_hash_v(unq_args)

    if(not source_in_cache(name, long_hash)): #or members.get('cache',True) == False):
        source = gen_cre_func_source(*gen_src_args, long_hash)
        source_to_cache(name, long_hash, source)

    # Update members with jitted 'call' and 'check'
    to_import = ['cf_type', 'call_heads'] + (['check'] if(has_check) else [])
    l = import_from_cached(name, long_hash, to_import)

    typ = l['cf_type']
    typ.call_heads = l['call_heads']
    if(has_check): typ.check = l['check']
    
    for k,v in members.items():
        setattr(typ,k,v)

    return typ


# -------------------------------------------------------------
# : CREFunc Source Generation

def gen_cre_func_source(name, return_type, arg_types, nopy_call, nopy_check,
             call_bytes, check_bytes, no_raise, ptr_args,
             long_hash ):
    '''Generate source code for the relevant functions of a user defined CREFunc.'''
    arg_names = ', '.join([f'a{i}' for i in range(len(arg_types))])

    on_error_map = {True : 'error', False : 'none', None : 'warn'}
    on_error_call = on_error_map[nopy_call]
    on_error_check = on_error_map[nopy_check]

    nl = "\n"
    source = \
f'''import numpy as np
from numba import njit, void, i8, u1, boolean, objmode
from numba.types import unicode_type
from numba.extending import lower_cast
from numba.core.errors import NumbaError, NumbaPerformanceWarning
from cre.utils import _incref_structref, cast, PrintElapse, _load_ptr, _obj_cast_codegen, _store_safe, _struct_get_attr_ptr
from cre.func import (ensure_repr_const, cre_func_ctor, CREFunc_method, CREFunc_assign_method_addr,
    CREFuncTypeClass, CREFuncType, VarType, CFSTATUS_TRUTHY, CFSTATUS_FALSEY, CFSTATUS_NULL_DEREF, CFSTATUS_ERROR)
from cre.fact import BaseFact, resolve_deref_data_ptr
import cloudpickle


return_type = cloudpickle.loads({cloudpickle.dumps(return_type)})
arg_types = cloudpickle.loads({cloudpickle.dumps(arg_types)})
cf_type = CREFuncTypeClass(return_type, arg_types,is_composed=False, name={name!r}, long_hash={long_hash!r})


@lower_cast(cf_type, CREFuncType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty, incref=False)

{"call_sig = return_type(*arg_types)" if not ptr_args else
 "call_sig = return_type(*([i8]*len(arg_types)))"}
call_pyfunc = cloudpickle.loads({call_bytes})
{"".join([f'h{i}_type, ' for i in range(len(arg_types))]) + " = arg_types" if(len(arg_types)) else ""}

call_heads = CREFunc_method(cf_type, call_sig, 'call_heads', on_error={on_error_call!r})(call_pyfunc)
if(call_heads is None):
    @CREFunc_method(cf_type, call_sig, 'call_heads')
    def call_heads({arg_names}):
        with objmode(_return=return_type):
            _return = call_pyfunc({arg_names})
        return _return

@CREFunc_method(cf_type, u1(CREFuncType))
def resolve_heads(_self):
    self = cast(_self, cf_type)
    has_null_deref = False
    {"".join([f"""
    if(self.root_arg_infos[{i}].has_deref):
        var = cast(self.ref{i}, VarType)
        a = cast(self.a{i}, BaseFact)
        data_ptr = resolve_deref_data_ptr(a, var.deref_infos)
        if(data_ptr != 0):
            self.h{i} = _load_ptr(h{i}_type, data_ptr)
        else:
            has_null_deref = True
""" for i in range(len(arg_types))])
    }
    if(has_null_deref):
        return CFSTATUS_NULL_DEREF
    return u1(0)

@CREFunc_method(cf_type, u1(CREFuncType))
def call_self(_self):
    self = cast(_self, cf_type)
'''
    if(no_raise):
        source +=f'''
    return_val = call_heads({", ".join([
            (f'_load_ptr(i8,_struct_get_attr_ptr(self,"h{i}"))' if(ptr_args) else f'self.h{i}') 
            for i in range(len(arg_types))])
        })
    _store_safe(return_type, self.return_data_ptr, return_val)
    return {"u1(return_val > 0)"
            if(isinstance(return_type, types.Number)) else (
             "CFSTATUS_TRUTHY if(return_val) else CFSTATUS_FALSEY"
             if(not isinstance(return_type, types.StructRef)) else
             "CFSTATUS_TRUTHY"
            )}
'''
    else:    
        source +=f'''
    try:
        return_val = call_heads({", ".join([
            (f'cast(self.h{i}, i8)' if(ptr_args) else f'self.h{i}')
            for i in range(len(arg_types))])
        })
        _store_safe(return_type, self.return_data_ptr, return_val)
        return {"u1(return_val > 0)"
                if(isinstance(return_type, types.Number)) else (
                 "CFSTATUS_TRUTHY if(return_val) else CFSTATUS_FALSEY"
                 if(not isinstance(return_type, types.StructRef)) else
                 "CFSTATUS_TRUTHY"
                )}
    except Exception:
        return CFSTATUS_ERROR    
'''
    if(check_bytes is not None):
        source += f'''
# check_pyfunc = cloudpickle.loads({check_bytes})
# check_sig = boolean(*arg_types)

# check = CREFunc_method(cf_type, check_sig, 'check', on_error={on_error_check!r})(check_pyfunc)
# if(check is None):
#     @CREFunc_method(cf_type, 'check', check_sig)
#     def check({arg_names}):
#         with objmode(_return=boolean):
#             _return = check_pyfunc({arg_names})
#         return _return
'''
    else:
        source += f'''
# CREFunc_assign_method_addr(cf_type, 'check', -1)
'''
    source += f'''
# @CREFunc_method(cf_type, boolean(*arg_types))
# def match({arg_names}):
#     {f"if(not check({arg_names})): return 0" if(check_bytes is not None) else ""}
#     return {f'1' if isinstance(return_type, StructRef) else f'1 if(call({arg_names})) else 0'}

# match_heads = match
# CREFunc_assign_method_addr(cf_type, 'match_heads', match.cre_method_addr)

# @CREFunc_method(cf_type, boolean(i8[::1],))
# def match_head_ptrs(ptrs):
# {indent(nl.join([f'#i{i} = _load_ptr(h{i}_type,ptrs[{i}])' for i in range(len(arg_types))]),prefix='    ')}
#     return match_heads({",".join([f'i{i}' for i in range(len(arg_types))])})


'''

    source +=f'''

# Make sure that constant members can be repr'ed
for at in arg_types:
    ensure_repr_const(at)

# ctor = gen_cre_func_ctor(cf_type)

@njit(CREFuncType(unicode_type, unicode_type, unicode_type), cache=True)
def ctor(name, expr_template, shorthand_template):
    return cre_func_ctor(cf_type, name, expr_template, shorthand_template, {"True" if ptr_args else "False"})
cf_type.ctor = ctor
'''

    return source

# ---------------------------------------------------------------------------
# : Unique String (independant of base variables)

@njit(unicode_type(CREFuncType), cache=True)
def cre_func_unique_string(self):
    '''Outputs a tuple that uniquely identifies an instance
         of a literal independant of the base Vars of its underlying CREFunc.
    '''
    stack = List()
    cf = self
    i = 0 
    arg_strs = List.empty_list(unicode_type)
    s = ""
    keep_looping = True
    while(keep_looping):
        arg_info = cf.root_arg_infos[i]
        if(arg_info.type == ARGINFO_FUNC):
            stack.append((cf, i+1,arg_strs))
            cf = cast(arg_info.ptr, CREFuncType)
            arg_strs = List.empty_list(unicode_type)
            i = 0 
        else:
            if(arg_info.type == ARGINFO_VAR):
                var = cast(arg_info.ptr, VarType)
                vs = ""
                for j, d in enumerate(var.deref_infos):
                    delim = "," if j != len(var.deref_infos)-1 else ""
                    vs += f'({str(i8(d.type))},{str(i8(d.t_id))},{str(i8(d.a_id))},{str(i8(d.offset))}){delim}'
                arg_strs.append(vs) 
            elif(arg_info.type == ARGINFO_CONST):
                addr = cf.origin_data.repr_const_addrs[i]
                fn = _func_from_address(rc_fn_typ, addr)
                arg_strs.append(fn(cf,i))
            i += 1
        while(i >= len(cf.root_arg_infos)):
            _a = ','.join([f'{{{str(i)}}}' for i in range(len(cf.root_arg_infos))])
            tmp = f"{str(cf.call_self_addr)}({_a})"
            s = tmp.format(arg_strs)
            if(len(stack) == 0):
                keep_looping = False
                break

            cf, i, arg_strs = stack.pop(-1)
            arg_strs.append(s)
    return s

# ----------------------------------------------------------------------------
# : UntypeCREFunc

def resolve_return_type(x):
    # Helper function for determining how to type CREFunc arguments
    if(isinstance(x,Var)):
        return x.head_type
    elif(isinstance(x,CREFunc)):
        return x.return_type
    elif(isinstance(x, UntypedCREFunc)):
        raise ValueError(f"Cannot resolve return type of UntypedCREFunc {x}")
    elif(isinstance(x, (int))):
        return types.int64
    elif(isinstance(x, (float))):
        return types.float64
    elif(isinstance(x, (str))):
        return types.unicode_type

class UntypedCREFunc():
    '''An CREFunc that has not been given a signature yet'''
    def __init__(self, name, members):
        self.func_name = self.name = name
        self.members = members
        self.members['root_arg_names'] = inspect.getfullargspec(self.members['call'])[0]
        self.root_arg_names = self.members['root_arg_names']
        self.members['call'] = njit(cache=True)(self.members['call'])
        self._specialize_cache = {}

    def __repr__(self):
        return f'UntypedCREFunc(name={self.name}, members={self.members})'

    def __str__(self):
        return f'{self.name}({", ".join(self.root_arg_names)})'


    def __call__(self,*args):
        if(len(self.root_arg_names) != len(args)):
            raise ValueError(f"Got {len(args)} args for UntypedCREFunc {self.name!r} with {len(self.root_arg_names)} positional arguments.")

        args, all_const, head_types = _standardize_py_args(self, args)

        # Cache singleton CREFunc definition for args of 'head_types'.
        if(head_types not in self._specialize_cache):
            call = self.members['call']
            if(not self.members.get("ptr_args", False)):
                sig = dispatcher_sig_from_arg_types(call, tuple(head_types))
            else:
                sig = dispatcher_sig_from_arg_types(call, tuple([i8]*len(head_types)))

            members = {**self.members, 'signature': sig}#
            cf = new_cre_func(self.name, members)
            self._specialize_cache[head_types] = cf

        # Pass 'args' to the singleton instance to either call or compose it.
        cf = self._specialize_cache[head_types]
        return cf(*args)

# ----------------------------------------------------------------------------
# : PtrCREFunc

# def define_PtrCREFunc(name, members):
#     ''' Defines a new CREFuncType '''

#     # Ensure that 'call' is present and that 'call' and 'check' are functions
#     assert 'call' in members, "PtrCREFunc must have call() defined"
#     assert hasattr(members['call'], '__call__'), "call() must be a function"

#     # Get pickle bytes for 'call_head_ptrs'
#     call_bytes, call_unq = _standardize_method(members, 'call')

#     # Regenerate the source for this type if wasn't cached or if 'cache=False' 
#     gen_src_args = [name, members['return_type'], call_bytes]
#     unq_args = [name, members['return_type'], call_unq]

#     long_hash = unique_hash(unq_args)

#     # print(name, long_hash)
#     if(not source_in_cache(name, long_hash)): #or members.get('cache',True) == False):
#         source = gen_ptr_cre_func_source(*gen_src_args, members['arg_types'], long_hash)
#         source_to_cache(name, long_hash, source)

#     # Update members with jitted 'call_head_ptrs'
#     to_import = ['cf_type', 'call']
#     l = import_from_cached(name, long_hash, to_import)

#     typ = l['cf_type']
#     typ.func_name = name
#     typ.call = l['call']
    
#     for k,v in members.items():
#         setattr(typ,k,v)

#     return typ

# def new_ptr_cre_func(name, members):
#     call = members['call']
#     arg_types = members['arg_types'] = tuple([i8]*len(members['arg_names']))

#     # Determine the signature 
#     if('signature' not in members):
#         if('return_type' in members):
#             members['signature'] = members['return_type'](arg_types)
#         else:
#             sig = dispatcher_sig_from_arg_types(call, arg_types)
#             members['signature'] = sig
#             members['return_type'] = sig.return_type

#     cf_type = define_PtrCREFunc(name, members)

#     expr_template = f"{name}({', '.join([f'{{{i}}}' for i in range(len(arg_types))])})"
#     shorthand_template = members.get('shorthand',expr_template)

#     cre_func = cf_type.ctor(
#         name,
#         expr_template,
#         shorthand_template                
#     )

#     cre_func._return_type = members['return_type']
#     cre_func._arg_types = arg_types
#     # cre_func.cf_type = cf_type
#     return cre_func

# class PtrCREFunc(CREFunc):

#     def __new__(self,*args, **kwargs):
#         ''' A decorator function that builds a new Func'''
#         # if(len(args) > 1): raise ValueError("PtrCREFunc() takes at most one position argument 'signature'.")
        
#         def wrapper(call_func):
#             # Make new cre_func_type
#             members = kwargs
#             members["call"] = njit(cache=True)(call_func)
#             name = call_func.__name__
#             members["arg_names"] = inspect.getfullargspec(call_func)[0]
#             # If no signature is given then create an untyped instance that can be specialized later
#             # if('signature' not in members):
#             #     return UntypedCREFunc(name,members)
#             # print(">>", members)
#             return new_ptr_cre_func(name, members)
            

#         if(len(args) == 1):
#             if(isinstance(args[0],(str, numba.core.typing.templates.Signature))):
#                 kwargs['signature'] = args[0]
#             elif(hasattr(args[0],'__call__')):
#                 return wrapper(args[0])
#             else:
#                 raise ValueError(f"Unrecognized type {type(args[0])} for 'signature'")        

#         return wrapper

#     def __call__(self,*py_args):
#         assert len(py_args) == self.nargs, f"{str(self)} takes {self.nargs}, but got {len(py_args)}"
#         if(any([not isinstance(x,Var) for x in py_args])):
#             raise ValueError("PtrOps only take Vars as input")
#             # If all of the arguments are constants then just call the CREFunc
#             # return self.call(*py_args)
#         else:
#             # head_var_ptrs = np.empty((self.nargs,),dtype=np.int64)
#             # for i, v in enumerate(py_args):
#             #     head_var_ptrs[i] = v.get_ptr()
#             return self.__class__.make_singleton_inst(py_args)



# def gen_ptr_cre_func_source(name, return_type, call_bytes, arg_types, long_hash):
#     source = \
# f'''import numpy as np
# from numba import njit, void, i8, boolean, objmode
# from numba.types import unicode_type
# from numba.extending import lower_cast
# from numba.core.errors import NumbaError, NumbaPerformanceWarning
# from cre.utils import PrintElapse, _struct_from_ptr, _load_ptr, _obj_cast_codegen, _store_safe, 
# from cre.func import ensure_repr_const, cre_func_ctor, CREFunc_method, CREFunc_assign_method_addr, CREFuncTypeClass, CREFuncType, VarType
# import cloudpickle

# return_type = cloudpickle.loads({cloudpickle.dumps(return_type)})
# arg_types = cloudpickle.loads({cloudpickle.dumps(arg_types)})

# cf_type = CREFuncTypeClass(return_type, arg_types, is_composed=False, name={name!r}, long_hash={long_hash!r})

# call_pyfunc = cloudpickle.loads({call_bytes})
# call_sig = return_type(*arg_types)
# call = CREFunc_method(cf_type, call_sig, 'call', on_error='error')(call_pyfunc)

# call_heads = call

# @CREFunc_method(cf_type, return_type(i8[::1]))
# def call_head_ptrs(ptrs):
#     return call_heads({",".join([f'ptrs[{i}]' for i in range(len(arg_types))])})

# @CREFunc_method(cf_type, boolean(i8[::1]))
# def match_head_ptrs(ptrs):
#     return {f'1' if isinstance(return_type, StructRef) else
#             f'1 if(call_head_ptrs(ptrs)) else 0'}

# #Placeholder repr
# ensure_repr_const(i8)

# @njit(CREFuncType(unicode_type, unicode_type, unicode_type), cache=True)
# def ctor(name, expr_template, shorthand_template):
#     return cre_func_ctor(cf_type, name, expr_template, shorthand_template, True)
# cf_type.ctor = ctor
# '''
#     return source



from numba import types, njit, generated_jit
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba.typed import List, Dict
from numba.core.types import DictType, ListType, unicode_type, Tuple
# from numba.cpython.unicode import  _set_code_point
from cre.core import (TYPE_ALIASES, DEFAULT_REGISTERED_TYPES, JITSTRUCTS,
                        py_type_map, numba_type_map, numpy_type_map,
                        unpickle_type_from_t_id, t_id_from_type_name,
                        add_to_type_registry, SHORT_NAMES, DEFAULT_TYPE_T_IDS)
from cre.caching import unique_hash
# from cre.struct_gen import gen_struct_code
from cre.structref import define_structref
from cre.utils import PrintElapse, _set_global, _get_global, cast, _incref_ptr
from numba.extending import overload_method
from numba.experimental.structref import new
import numpy as np
import os

t_id_a_id_tup_type = Tuple((u2,u1))
context_data_fields = {
    "name": unicode_type,
    "unhandled_retro_registers": ListType(u2),
    "fact_to_t_id": DictType(unicode_type, u2),
    "t_id_to_type_names": DictType(u2, unicode_type),
    "parent_t_ids": ListType(u2[::1]),
    "child_t_ids": ListType(u2[::1]),
    "attr_names" : DictType(t_id_a_id_tup_type, unicode_type)
}

CREContextData, CREContextDataType, CREContextDataTypeClass  = define_structref("CREContextData",context_data_fields, define_constructor=False, return_type_class=True)

@njit(types.void(CREContextDataType), cache=True)
def set_cre_context_data(context_data):
    _set_global(i8, "_CRE_context_data", cast(context_data,i8))

@njit(CREContextDataType(), cache=True)
def get_cre_context_data():
    context_data_ptr = _get_global(i8, "_CRE_context_data")
    _incref_ptr(context_data_ptr)
    return cast(context_data_ptr, CREContextDataType)

u2_arr_type = u2[::1]
@njit(CREContextDataType(unicode_type), cache=True)
def new_cre_context(name):
    st = new(CREContextDataType)
    st.name = name
    st.unhandled_retro_registers = List.empty_list(u2)
    st.fact_to_t_id = Dict.empty(unicode_type, u2)
    st.t_id_to_type_names = Dict.empty(u2, unicode_type)
    st.parent_t_ids = List.empty_list(u2_arr_type)
    st.child_t_ids = List.empty_list(u2_arr_type)
    st.attr_names = Dict.empty(t_id_a_id_tup_type, unicode_type)
    return st 


@njit(types.void(CREContextDataType, unicode_type, u2), cache=True)
def assign_name_to_t_id(cd, name, t_id):
    cd.fact_to_t_id[name] = u2(t_id)
    cd.t_id_to_type_names[u2(t_id)] = name

@njit(types.void(CREContextDataType, u2, u1, unicode_type), cache=True)
def assign_a_id_attr(cd, t_id, a_id, name):
    cd.attr_names[(u2(t_id), u1(a_id))] = name 

@njit(cache=True)
def ensure_inheritance(cd, t_id, inh_t_id=-1):
    # Ensure that parent_t_ids and child_t_ids are big enough
    for i in range(len(cd.parent_t_ids),t_id+1):
        cd.parent_t_ids.append(np.zeros((0,),dtype=np.uint16))

    for i in range(len(cd.child_t_ids),t_id+1):
        cd.child_t_ids.append(np.zeros((0,),dtype=np.uint16))

    # Use inh_fact_num to fill in the parents 
    if(inh_t_id != -1):
        old_arr = cd.parent_t_ids[inh_t_id]
        new_arr = np.empty((len(old_arr)+1,),dtype=np.uint16)
        new_arr[:len(old_arr)] = old_arr
        new_arr[-1] = inh_t_id
        cd.parent_t_ids[t_id] = new_arr
    
    # Use the updated parents to update child relations (facts count as their own child) 
    for p_t_id in cd.parent_t_ids[t_id]:
        if(t_id not in cd.child_t_ids[p_t_id]):
            old_arr = cd.child_t_ids[p_t_id]
            new_arr = np.empty((len(old_arr)+1,),dtype=np.uint16)
            new_arr[:len(old_arr)] = old_arr
            new_arr[-1] = t_id
            cd.child_t_ids[p_t_id] = new_arr

    # Always treat as own child
    if(t_id not in cd.child_t_ids[t_id]):
        old_arr = cd.child_t_ids[t_id]
        new_arr = np.empty((len(old_arr)+1,),dtype=np.uint16)
        new_arr[:len(old_arr)] = old_arr
        new_arr[-1] = t_id
        cd.child_t_ids[t_id] = new_arr



@njit(cache=True)
def clear_unhandled_retro_registers(self):
    self.unhandled_retro_registers = List.empty_list(u2)

@njit(cache=True)
def get_unhandled_retro_registers(self):
    arr = np.empty((len(self.unhandled_retro_registers),), dtype=np.uint16)
    for i, t_id in enumerate(self.unhandled_retro_registers):
        arr[i] = t_id
    return arr

@njit(cache=True)
def has_unhandled_retro_registers(self):
    return len(self.unhandled_retro_registers) > 0

class CREContext(object):
    _contexts = {}

    @classmethod
    def get_default_context(cls):
        '''Returns the default context set in the cre_DEFAULT_CONTEXT 
            environment variable.'''
        df_c = os.environ.get("cre_DEFAULT_CONTEXT")
        df_c = df_c if df_c else "cre_default"
        return cls.get_context(df_c)

    @classmethod
    def init(cls, name):
        ''' Builds a new context with 'name'.'''
        if(name not in cls._contexts):
            self = cls(name)
            self.enter_count = 0
            cls._contexts[name] = self
        else:
            raise ValueError(f"Context redefinition attempted for name {name}.")

    @classmethod
    def get_context(cls, name=None):
        ''' Gets a context by name or the current scope's context if no name
            is given. Will instantiate a new context if no context with of 'name' exists.'''
        if(name is None):
            return cls.get_context(cre_context_ctxvar.get(cls.get_default_context()))
        if(isinstance(name,CREContext)): return name
        if(name not in cls._contexts): cls.init(name)
        return cls._contexts[name]

    @classmethod
    def set_default_context(cls, name):
        os.environ["cre_DEFAULT_CONTEXT"] = name

    def __init__(self,name):
        self.name = name
        self.name_to_type = {**DEFAULT_REGISTERED_TYPES}
        self.name_to_type.update({str(x): x for x in DEFAULT_REGISTERED_TYPES.values()})

        self.t_id_to_type = list(DEFAULT_REGISTERED_TYPES.values())
        self.type_to_t_id = {typ:i for i,typ in enumerate(self.t_id_to_type)}

        self.op_instances = {}
        self.deferred_types = {}
        
        self.parents_of = {}
        self.children_of = {}
        
        self.context_data = cd = new_cre_context(name)

        for s,t in DEFAULT_REGISTERED_TYPES.items():
            t_id = DEFAULT_TYPE_T_IDS[s]
            s = SHORT_NAMES.get(t, s)
            assign_name_to_t_id(self.context_data, s, t_id)

        
    def get_deferred_type(self,name):
        if(name not in self.deferred_types):
            from cre.fact import DeferredFactRefType
            self.deferred_types[name] = DeferredFactRefType(name)
        return self.deferred_types[name]


    def _ensure_retro_registers(self):
        from cre.core import DEFAULT_REGISTERED_TYPES

        cd = self.context_data
        if(has_unhandled_retro_registers(cd)):
            child_t_ids = cd.child_t_ids
            for t_id in get_unhandled_retro_registers(cd):
                if(len(child_t_ids[t_id]) == 0):
                # if(t_id >= len(self.t_id_to_type) or self.t_id_to_type[t_id] is None):
                    self._retroactive_register(t_id)
            clear_unhandled_retro_registers(cd)

    @property
    def registered_types(self):
        self._ensure_retro_registers()
        return self.t_id_to_type

    @property
    def unhandled_retro_registers(self):
        return get_unhandled_retro_registers(self.context_data)

    def _assert_written_to_type_registry(self, typ):
        '''Ensures that a type is written to the on disk type registery'''
        if(typ not in self.type_to_t_id):
            hash_code = getattr(typ,'_hash_code',unique_hash(typ.name))
            t_id = t_id_from_type_name(str(typ), hash_code)
            name = str(typ)
            if(t_id == -1):
                t_id = add_to_type_registry(name, hash_code, typ)
            self._assign_name_t_id(str(typ), typ, t_id)

    def _assign_name_t_id(self, name, typ, t_id):
        '''For this context associate a name and t_id with typ.''' 
        assign_name_to_t_id(self.context_data, name, t_id)

        # Fill in the python facing 'name_to_type'
        # print(name, "->", str(typ))
        # if(str(typ) == "TextField_vVBHx8mvD8"):
        #     raise ValueError()
        self.name_to_type[name] = typ
        self.name_to_type[str(typ)] = typ

        # Ensure 't_id_to_type' is long enough. In the case of a retroactive
        #  registration there can be holes so fill with None first then assign t_id
        for i in range(len(self.t_id_to_type),t_id+1):
            self.t_id_to_type.append(None)
        self.t_id_to_type[t_id] = typ
        self.type_to_t_id[typ] = t_id

    def _register_attr_names(self, typ, t_id):
        for a_id, attr in enumerate(typ.field_dict_keys):
            assign_a_id_attr(self.context_data, u2(t_id), u1(a_id), attr)

    def _register_fact_type(self, name, fact_type, inherit_from=None):
        '''Registers a fact_type to this context. Keeps track of inheritance and casting information'''

        t_id = fact_type.t_id
        inh_t_id = inherit_from.t_id if inherit_from is not None else -1
        ensure_inheritance(self.context_data, t_id, inh_t_id)
        self._assign_name_t_id(name, fact_type, t_id)
        self._register_attr_names(fact_type, t_id)


        
        # Dispatcher args of type inherit_from also accept fact_type 
        from numba.core.typeconv.rules import TypeCastingRules, default_type_manager as tm
        if(inherit_from): 
            tm.set_safe_convert(fact_type, inherit_from)
            tm.set_safe_convert(types.ListType(fact_type), types.ListType(inherit_from))

        # Dispatcher args of type BaseFact also accept fact_type.
        from cre.fact import BaseFact
        tm.set_safe_convert(fact_type, BaseFact)
        tm.set_safe_convert(types.ListType(fact_type), types.ListType(BaseFact))
        
        # Track inheritence structure
        i_name = inherit_from._fact_name if inherit_from else None
        self.parents_of[name] = self.parents_of.get(i_name,[]) + [inherit_from] if i_name else []
        if(i_name):
            for parent in self.parents_of[name]:
                p = parent._fact_name
                self.children_of[p] = self.children_of.get(p,[]) + [fact_type]
                self.children_of[parent] = self.children_of[p]
        self.children_of[name] = []

        # Index inheritance on both the type and it's name
        self.parents_of[fact_type] = self.parents_of[name]
        self.children_of[fact_type] = self.children_of[name]

        
    def get_type(self, name:str = None, t_id:int = None, ensure_retro=True):
        '''Retrieves the type associated with a user defined fact given a
            name, fact_num, or t_id. '''
        if(ensure_retro): self._ensure_retro_registers()
        
        # If t_id defined then retrieve the type for t_id
        if(t_id is not None):
            typ = None
            if(t_id < len(self.t_id_to_type)):
                typ = self.t_id_to_type[t_id]
                if(typ is types.undefined): typ = None
            if(typ is None): typ = self._retroactive_register(t_id)
            if(typ is None): 
                print(f"No type with t_id={t_id} registered in cre_context {self.name}.")
                raise ValueError(f"No type with t_id={t_id} registered in cre_context {self.name}.")
            return typ
        
        # If got a name then check the registry for the name 
        if(name is not None):
            if(not isinstance(name, str)):
                raise ValueError(f"Type name should be str got {type(name)}")
            if(name in self.name_to_type):
                # print("get_type->", name, self.name_to_type[name], "in", self.name)
                return self.name_to_type[name]
            else:
                raise ValueError(f"No type {name} registered in cre_context {self.name}.")
                
        raise ValueError("Bad arguments for 'get_type'. Expecting one keyword argument name:str, or t_id:int")

    def _retroactive_register(self, t_id):
        '''Retroactively registers a type in cases where a fact_type is used but not defined (e.g. if cached dispatcher 
            returns an instance of it). Pulls the type definition from hard drive cache to register it retroactively.'''
        if(t_id == 0): 
            raise ValueError("Tried to register t_id=0 (i.e. undefined)")
        try:
            ft = ret_typ = _type = unpickle_type_from_t_id(t_id)
        except ImportError:
            return

        types = []
        while(ft is not None):
            types.append(ft)
            ft = getattr(ft,"parent_type",None)
        for ft in reversed(types):
            if(hasattr(ft,"_fact_name")):
                self._register_fact_type(ft._fact_name, ft, getattr(ft,'parent_type', None))
            else:
                self._assign_name_t_id(str(ft), ft, t_id)
        return ret_typ

    def get_t_id(self, _type=None, name:str=None, retro_register:bool=True):
        '''Retreives the t_id for a _type, or the name of the type'''
        if(name is not None):
            _type = self.get_type(name=name)

        if(_type is not None):
            if(hasattr(_type,'t_id')):
                t_id = _type.t_id
            else:
                self._assert_written_to_type_registry(_type)
                t_id = self.type_to_t_id[_type]


            return t_id

        raise ValueError("Bad arguments for 'get_t_id'. Expecting one keyword argument name:str, or _type:Type")

    
    def get_parent_t_ids(self, _type=None, name:str=None, t_id=None):
        if(t_id is None):
            t_id = self.get_t_id(_type=_type,name=name)

        if(t_id is None):
            raise ValueError("Bad arguments for 'get_parent_t_ids'. Expecting one keyword argument name:str, t_id:int, or _type:Type")

        p_t_ids = self.context_data.parent_t_ids
        if(t_id >= len(p_t_ids) or len(p_t_ids[t_id])):
            self._retroactive_register(t_id)
        p_t_ids = self.context_data.parent_t_ids

        return p_t_ids[t_id]
        

    def get_child_t_ids(self, _type=None, name:str=None, t_id=None, inclusive=True):
        # NOTE: Fix inclusive
        if(t_id is None):
            t_id = self.get_t_id(_type=_type,name=name)

        if(t_id is None):
            raise ValueError("Bad arguments for 'get_parent_t_ids'. Expecting one keyword argument name:str, t_id:int, or _type:Type")

        c_t_ids = self.context_data.child_t_ids
        if(t_id >= len(c_t_ids) or len(c_t_ids[t_id])):
            self._retroactive_register(t_id)
        c_t_ids = self.context_data.child_t_ids

        return c_t_ids[t_id]#[0 if inclusive else 1:]

    def __str__(self):
        return f"CREContext({self.name})"

    def __enter__(self):
        self.enter_count += 1
        self.token_prev_context = cre_context_ctxvar.set(self.name)
        set_cre_context_data(self.context_data)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.enter_count -= 1
        if(self.enter_count == 0):
            cre_context_ctxvar.reset(self.token_prev_context)
            set_cre_context_data(CREContext.get_context().context_data)
        if(exc_val): raise exc_val.with_traceback(exc_tb)

        return self
    
def cre_context(context=None):
    return CREContext.get_context(context)

set_cre_context_data(CREContext.get_default_context().context_data)



import numba 
import numpy as np
import numba
from numba import types, jit,njit, generated_jit
from numba import deferred_type, optional
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba.typed import List, Dict
from numba.types import ListType, unicode_type, UnicodeType
from numba.cpython.unicode import _empty_string, _set_code_point, _get_code_point, PY_UNICODE_1BYTE_KIND
from numba.extending import overload, overload_method
# from numba.cpython.unicode import _empty_string, _set_code_point, _get_code_point, PY_UNICODE_1BYTE_KIND


DIGITS_START = 48
DIGITS_END = 58
DASH = 45
DOT = 46
PLUS = 43
E_CHAR = 101

## FROM STRING ##

# @njit(i8(unicode_type,),cache=True)
@njit(cache=True)
def str_to_posint(s):
    final_index, result = len(s) - 1, 0
    for i,v in enumerate(s):
        char = ord(v)
        if(char < DIGITS_START or char >= DIGITS_END):
            raise ValueError("Could not convert string to numeric type.")
        result += (char - 48) * (10 ** (final_index - i))
    return result

# @njit(i8(unicode_type,), cache=True)
@njit(cache=True)
def str_to_int(s):
    neg = (s[0] == "-")
    if(neg): s = s[1:]
    result = str_to_posint(s)
    return -result if neg else result

@njit(cache=True)
def str_to_float(s):
    neg = (s[0] == "-")
    if(neg): s = s[1:]

    exp, dec_loc, exp_loc = 1,-1, len(s)
    for i,c in enumerate(s):
        if(c == "."): 
            dec_loc=i
        elif(c == "e"):
            exp_loc=i
            exp = str_to_int(s[exp_loc+1:])

    if(dec_loc != -1):
        result = str_to_posint(s[:dec_loc])
        result += float(str_to_posint(s[dec_loc+1:exp_loc])) / np.power(10,(exp_loc-dec_loc-1))
    else:
        result = str_to_posint(s)

    if(exp != 1):
        result *= 10.**exp

    if(neg): result = -result
    return result


## TO STRING ##

@njit(cache=True)
def int_to_str(x):
    isneg = int(x < 0.0)
    x = np.abs(x)
    l = 0 
    _x = x
    while _x > 0:
        _x = _x // 10
        l += 1
    s = _empty_string(PY_UNICODE_1BYTE_KIND,l+isneg)
    if(isneg): _set_code_point(s,0,DASH)
    for i in range(l):
        digit = x % 10
        _set_code_point(s,isneg+l-i-1,digit + DIGITS_START)
        x = x // 10
    return s

# if(_x < 1e-128): _x = (_x * 1e128) % 10; l2 += 128; print(_x,"??128")
    # if(_x < 1e-64): _x = (_x * 1e64) % 10; l2 += 64; print(_x,"??64")
    # if(_x < 1e-32): _x = (_x * 1e32) % 10; l2 += 32; print(_x,"??32")
    # if(_x < 1e-16): _x = (_x * 1e16) % 10; l2 += 16; print(_x,"??16")
    # if(_x < 1e-8): _x = (_x * 1e8) % 10; l2 += 8; print(_x,"??8")
    # if(_x < 1e-4): _x = (_x * 1e4) % 10; l2 += 4; print(_x,"??4")
    # if(_x < 1e-2): _x = (_x * 1e2) % 10; l2 += 2; print(_x,"??2")
    # if(_x < 1e-1): _x = (_x * 1e1) % 10; l2 += 1; print(_x,"??1")
    

@njit(cache=True)
def get_n_digits(x):
    l1,l2 = 0,-1
    _x = x
    while _x > 0:
        _x = _x // 10
        l1 += 1

    _x = x % 10     
    while _x > 1e-10:
        _x = (_x * 10) % 10
        l2 += 1
        if(l2 >= 16): break
    return l1, l2


@njit(cache=True)
def float_to_str(x):
    if(x == np.inf):
        return 'inf'
    elif(x == -np.inf):
        return '-inf'

    isneg = int(x < 0.0)
    x = np.abs(x)
    
    if(x != 0.0):
        # There is probably a more efficient way to do this
        e = np.floor(np.log10(x))
        if(10**e - x > 0): e -= 1
    else:
        e = 0
    
    is_exp, is_neg_exp = e >= 16, e <= -16

    exp_chars = 0
    if (is_exp or is_neg_exp):
        exp_chars = 4
        if(e >= 100 or e <= -100): exp_chars = 5


    if(is_exp):
        offset_x = np.around(x * (10.0**-(e)),15)
        l1, l2 = get_n_digits(offset_x)
    elif(is_neg_exp):
        offset_x = np.around(x * (10**-(e)),15)
        l1, l2 = get_n_digits(offset_x)
    else:
        offset_x = x
        l1,l2 = get_n_digits(x)
        l2 = max(1,l2) # Will have at least .0 
    
    use_dec = l2 > 0

    # print("<<", e, offset_x, l2)

    l = l1+l2+use_dec
    length = l+isneg+exp_chars
    s = _empty_string(PY_UNICODE_1BYTE_KIND,length)
    if(isneg): _set_code_point(s,0,DASH)

    _x = offset_x
    for i in range(l1):
        digit = int(_x % 10)
        _set_code_point(s,(isneg+l1)-i-1,digit + DIGITS_START)
        _x = _x // 10

    if(use_dec):
        _set_code_point(s,l1+isneg,DOT)

    _x = offset_x % 10
    for i in range(l2):
        _x = (_x * 10) % 10
        digit = int(_x)  
        
        _set_code_point(s,(isneg+l1)+i+use_dec,digit + DIGITS_START)

    if(is_exp or is_neg_exp):
        i = (isneg+l1+use_dec+l2)
        _set_code_point(s,i,E_CHAR)        
        if(is_exp):
            _set_code_point(s,i+1,PLUS)     
        if(is_neg_exp):
            _set_code_point(s,i+1,DASH)           

        i = length-1
        exp = np.abs(e)
        while(exp > 0):
            digit = exp % 10
            _set_code_point(s,i,digit + DIGITS_START)
            exp = exp // 10
            i -= 1

    return s

# Monkey Patch type conversions so that float(str) and str(float)
# work like in normal python

@overload_method(types.Float, '__str__')
def overload_float_to_str(x):
    def impl(x):
        return float_to_str(x)
    return impl
# def overload_float_to_str(x):
#     if(x is types.bool_):
#         def impl(x):
#             return "True" if x else "False"
#     elif(x in types.real_domain): 
#         def impl(x):
#             return float_to_str(x)
#     elif(x in types.integer_domain):
#         def impl(x):
#             return int_to_str(x)
#     else:
#         return
#     return impl

@overload(float)
def overload_str_to_float(x):
    if(x != unicode_type): return
    def impl(x):
        return str_to_float(x)
    return impl

@overload(int)
def overload_str_to_int(x):
    if(x != unicode_type): return
    def impl(x):
        return str_to_int(x)
    return impl


@njit(cache=True,inline="never")
def format_str(s, args):
    strs = List.empty_list(unicode_type)
    start = 0
    end = 0
    for i,c in enumerate(s):
        if(c == "{"):
            end = i 
            strs.append(s[start:end])
        elif(c == "}"):
            start = i+1
            ind = int(s[end+1:i])
            if(ind >= len(args)): raise ValueError()
            strs.append(args[ind])

    strs.append(s[start:])
    return "".join(strs)

@generated_jit(cache=True,nopython=True)
@overload_method(UnicodeType,'format')
def overload_format(s, *args):
    zero_type = args[0]
    if(isinstance(zero_type, types.BaseTuple)):
        zero_type = zero_type[0]
    if(isinstance(zero_type,types.ListType)):
        def impl(s, *args):
            return format_str(s,args[0])
    else:
        def impl(s, *args):
            return format_str(s,args)
    return impl



from numba.core.typing.templates import (AttributeTemplate, ConcreteTemplate,
                                         AbstractTemplate, infer_global, infer,
                                         infer_getattr, signature,
                                         bound_function, make_callable_template)


@infer_global(float)
class Float(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws

        [arg] = args

        if(arg == unicode_type):
            #For unicode_type fall back on @overload 
            return 

        if arg not in types.number_domain:
            raise TypeError("float() only support for numbers")

        if arg in types.complex_domain:
            raise TypeError("float() does not support complex")

        if arg in types.integer_domain:
            return signature(types.float64, arg)

        elif arg in types.real_domain:
            return signature(arg, arg)

base64 = ("_","A","B","C","D","E","F","G","H","I","J","K","L","M","N","O",
              "P","Q","R","S","T","U","V","W","X","Y","Z","a","b","c","d",
              "e","f","g","h","i","j","k","l","m","n","o","p","q","r","s",
              "t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9","_")
# for i,x in enumerate(base64):
#     print(i,x)
base64_ord = tuple([u1(ord(x)) for x in base64])
@njit(unicode_type(i8),cache=True)
def ptr_to_var_name(ptr):
    x = u8(ptr)

    # Ignore first 24 bits, no one has a terabyte of RAM
    a = np.empty(5,dtype=np.uint8)
    a[0] = (x>>34) & u8(63)
    a[1] = (x>>28) & u8(63)
    a[2] = (x>>22) & u8(63)
    a[3] = (x>>16) & u8(63)
    a[4] = (x>>10) & u8(63)
    # Ignore last 4 bits, probably won't be used because of byte alignment
    
    tail = u1(0)
    # Ignore zeros in highest bits
    l_ind = 0
    if(a[0] == 0): l_ind += 1
    if(a[1] == 0): l_ind += 1

    # If the tailing char would be a number then set 'tail' char to 110000
    if(a[4] > 52):
        tail = u1(48)
        a[4] = a[4] & 15

    # Build the string
    s = _empty_string(PY_UNICODE_1BYTE_KIND,5-l_ind+1)
    c = 0
    for i in range(4,l_ind-1,-1):
        # Since there are only 63 legal characters to us in var names
        #   if a[i] is 63 modify the tail char.
        if(a[i] == 63):
            tail |= 1<<(c)
            _set_code_point(s,c,base64_ord[a[i]&31])
        else:
            _set_code_point(s,c,base64_ord[a[i]])
        c += 1

    if(tail != 0):
        _set_code_point(s,5-l_ind,base64_ord[tail])
    else:
        s = s[:-1]

    return s


base64_alpha_num = (
              "A","B","C","D","E","F","G","H","I","J","K","L","M","N","O",
              "P","Q","R","S","T","U","V","W","X","Y","Z","a","b","c","d",
              "e","f","g","h","i","j","k","l","m","n","o","p","q","r","s",
              "t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9","A","B")
base64_alpha_num_ord = tuple([u1(ord(x)) for x in base64_alpha_num])

@njit(unicode_type(u1[::1], i8), cache=True)
def byte_string_to_base64(byte_arr, n_char):
    s = _empty_string(PY_UNICODE_1BYTE_KIND, n_char)
    for i in range(n_char):
        bit_ind = (i*6)
        lower_ind = bit_ind >> 3
        upper_ind = (bit_ind+6) >> 3
        offset = max(bit_ind-(lower_ind << 3),0)
        overflow = max((bit_ind+6)-((lower_ind+1) << 3),0)
        lower_width = (6-overflow)

        byte = (byte_arr[lower_ind] >> offset) & (255 >> (8-lower_width))
        overflow_byte = byte_arr[upper_ind] & (255 >> (8-overflow))
        byte |= overflow_byte << lower_width
        
        _set_code_point(s, i, base64_alpha_num_ord[byte])
    return s


# @njit
# def failed_int():
#   int("1.2")

# print(repr(failed_str()))
# print(repr(failed_float()))
# failed_int()



from numba.typed import Dict, List
from numba.types import DictType, ListType
from numba import i8, types
from cre.caching import unique_hash_v, source_in_cache, import_from_cached, source_to_cache
from numba.experimental import structref
from numba.core.typeconv import Conversion
from numba.extending import intrinsic
from llvmlite import ir as llvmir
from numba.core import cgutils, errors, imputils, types, utils

def _gen_getter_jit(typ,attr):
    return f'''@njit(cache=True)
def {typ}_get_{attr}(self):
    return self.{attr}
'''

def _gen_getter(typ,attr):
    return f'''    @property
    def {attr}(self):
        return {typ}_get_{attr}(self)
    '''
    
def gen_structref_code(typ,fields,
    extra_imports="from numba.experimental import structref",
    register_decorator="@structref.register",
    define_constructor=True,
    define_boxing=True

    ):
    attrs = [x[0] if isinstance(x,tuple) else x for x in fields]
    getters = "\n".join([_gen_getter(typ,attr) for attr in attrs])
    getter_jits = "\n".join([_gen_getter_jit(typ,attr) for attr in attrs])
    attr_list = ",".join(["'%s'"%attr for attr in attrs])
    code = \
f'''
from numba.core import types
from numba import njit
from cre.structref import CastFriendlyStructref, define_boxing
{extra_imports}
{getter_jits}
{register_decorator}
class {typ}TypeTemplate(CastFriendlyStructref):
    pass
    # def preprocess_fields(self, fields):
    #     return tuple((name, types.unliteral(typ)) for name, typ in fields)

class {typ}(structref.StructRefProxy):
    def __new__(cls, *args):
        return structref.StructRefProxy.__new__(cls, *args)

{getters}

{f'structref.define_constructor({typ}, {typ}TypeTemplate, [{attr_list}])' if(define_constructor) else ''}
{f'define_boxing({typ}TypeTemplate, {typ})' if(define_boxing) else ''}


'''
    return code

# def gen_struct(typ,fields):
#     code = gen_structref_code(typ,fields)
#     print(code)
#     l = {}
#     exec(code,{},l)
#     print(l[f"{typ}TypeTemplate"])
#     return l[f"{typ}TypeTemplate"](fields=fields)

def define_structref_template(name, fields, define_constructor=True,define_boxing=True):
    if(isinstance(fields,dict)): fields = [(k,v) for k,v in fields.items()]
    hash_code = unique_hash_v([name,fields, define_constructor, define_boxing])
    # if(name == "ExplanationTreeEntry"): print(name, fields, hash_code)
    
    if(not source_in_cache(name,hash_code)):
        source = gen_structref_code(name, fields, define_constructor=define_constructor,
             define_boxing=define_boxing)
        source_to_cache(name,hash_code,source)
        
    ctor, type_class = import_from_cached(name,hash_code,[name,f"{name}TypeTemplate"]).values()
    ctor._hash_code = hash_code
    return ctor, type_class

def define_structref(name, fields, define_constructor=True, define_boxing=True, return_type_class=False):
    if(isinstance(fields,dict)): fields = [(k,v) for k,v in fields.items()]
    ctor, type_class = define_structref_template(name,fields, define_constructor=define_constructor,define_boxing=define_boxing)
    struct_type = type_class(fields=fields)
    struct_type._hash_code = ctor._hash_code
    if(return_type_class):
        return ctor, struct_type, type_class
    else:
        return ctor, struct_type


def define_boxing(struct_type, obj_class):
    '''Same as in numba.experimental.structref but give the type a reference to the proxy'''
    struct_type._proxy_class = obj_class
    structref.define_boxing(struct_type, obj_class)

class CastFriendlyStructref(types.StructRef):
    def can_convert_to(self, typingctx, other):
        """
        Convert this Record to the *other*.
        This method only implements width subtyping for records.
        """
        from numba.core.errors import NumbaExperimentalFeatureWarning
        if isinstance(other, CastFriendlyStructref):
            if len(other._fields) > len(self._fields):
                return
            for other_fd, self_fd in zip(other._fields,
                                         self._fields):
                if not other_fd == self_fd and other_fd[1] != types.Any and self_fd[1] != types.Any:
                    return
            # warnings.warn(f"{self} has been considered a subtype of {other} "
            #               f" This is an experimental feature.",
            #               category=NumbaExperimentalFeatureWarning)
            return Conversion.safe

    def can_convert_from(self, typingctx, other):
        return other.can_convert_to(typingctx, other)

    def unify(self, typingctx, other):
        if(self.can_convert_to(typingctx, other)):
            return other

    def hrepr(self):
        '''A human-readable repr for cre objects'''
        return f"{self._typename}({', '.join([f'{k}={str(v)}' for k,v in self._fields])})" 
# print(DictType(i8,i8))
# print(fields)
# print(gen_structref_code("BOOP",fields))


# fields = [('dict', DictType(i8,i8)),
#     ('v', i8)]

# gen_struct("BOOP", fields)


from numba.core.runtime.nrtdynmod import _meminfo_struct_type

def imp_dtor(context, module, instance_type, user_dtor=None):
    llvoidptr = context.get_value_type(types.voidptr)
    llsize = context.get_value_type(types.uintp)
    dtor_ftype = llvmir.FunctionType(llvmir.VoidType(),
                                     [llvoidptr, llsize, llvoidptr])

    
    # print(user_dtor)


    fname = "_Dtor.{0}".format(instance_type.name)
    dtor_fn = cgutils.get_or_insert_function(module, dtor_ftype, fname)
    if dtor_fn.is_declaration:
        # Define
        builder = llvmir.IRBuilder(dtor_fn.append_basic_block())

        alloc_fe_type = instance_type.get_data_type()
        alloc_type = context.get_value_type(alloc_fe_type)
        ptr = builder.bitcast(dtor_fn.args[0], alloc_type.as_pointer())

        if(True):
            
            # print(user_dtor_res.fndesc.__dict__)
            # print(user_dtor)

            alloc_size = context.get_abi_sizeof(alloc_type)
            mi_size = context.get_abi_sizeof(_meminfo_struct_type)
            # print(mi_size, type(mi_size))

            raw_data_ptr = builder.ptrtoint(dtor_fn.args[0], cgutils.intp_t)
            # A bit of a hack get meminfo from data pointer by abusing fact that it is allocated in the same block as its data 
            # raw_mi_ptr = builder.sub(raw_data_ptr, context.get_constant(types.intp, mi_size))
            # meminfo = builder.inttoptr(raw_mi_ptr, cgutils.voidptr_t)
            # raw_mi_ptr = context.cast(builder, raw_mi_ptr, types.intp, types.intp)
            # print(raw_mi_ptr.__dict__)

            # inst_struct_type = cgutils.create_struct_proxy(instance_type)
            # inst_struct = context.make_helper(builder, instance_type)
            # inst_struct.meminfo = meminfo

            # context.compile_internal(builder, user_dtor_py_func, types.void(types.intp,), (raw_mi_ptr,))

            # builder.call(user_dtor, [inst_struct._getvalue()])
            # builder.call(user_dtor, [inst_struct._getvalue()])
            builder.call(user_dtor, [raw_data_ptr])

        
        data = context.make_helper(builder, alloc_fe_type, ref=ptr)

        # print(context.nrt.get_meminfos(builder, alloc_fe_type, data._getvalue()))

        
        context.nrt.decref(builder, alloc_fe_type, data._getvalue())


        builder.ret_void()

    return dtor_fn



@intrinsic
def new(typingctx, struct_type, user_dtor_type=None):
    """new(struct_type)

    A jit-code only intrinsic. Used to allocate an **empty** mutable struct.
    The fields are zero-initialized and must be set manually after calling
    the function.

    Example:

        instance = new(MyStruct)
        instance.field = field_value
    """
    # from numba.experimental.jitclass.base import imp_dtor
    # print("<<", user_dtor.dispatcher.py_func)
    
    inst_type = struct_type.instance_type
    user_dtor_res = user_dtor_type.dispatcher.get_compile_result(types.void(i8,))
    # user_dtor_py_func = user_dtor_type.dispatcher.py_func

    # user_dtor = user_dtor_type.dispatcher.get_compile_result(types.void(inst_type)).entry_point
    # print(user_dtor_res)

    def codegen(context, builder, signature, args):
        # FIXME: mostly the same as jitclass ctor_impl()
        model = context.data_model_manager[inst_type.get_data_type()]
        alloc_type = model.get_value_type()
        alloc_size = context.get_abi_sizeof(alloc_type)

        # print("args", args)
        # user_dtor = args[1] if len(args) == 2 else None

        
        user_dtor = context.declare_external_function(builder.module, user_dtor_res.fndesc)
        meminfo = context.nrt.meminfo_alloc_dtor(
            builder,
            context.get_constant(types.uintp, alloc_size),
            imp_dtor(context, builder.module, inst_type, user_dtor),
        )
        data_pointer = context.nrt.meminfo_data(builder, meminfo)
        data_pointer = builder.bitcast(data_pointer, alloc_type.as_pointer())

        # user_dtor = context.declare_external_function(builder.module, user_dtor_res.fndesc)
        builder.call(user_dtor, [builder.ptrtoint(data_pointer, cgutils.intp_t)])


        # Nullify all data
        builder.store(cgutils.get_null_value(alloc_type), data_pointer)

        inst_struct = context.make_helper(builder, inst_type)
        inst_struct.meminfo = meminfo

        return inst_struct._getvalue()

    sig = inst_type(struct_type, user_dtor_type)
    return sig, codegen



_, StructRefType = define_structref("StructRef",[])



from cre.transform.memset_builder import MemSetBuilder
from cre.transform.enumerizer import Enumerizer
from cre.transform.flattener import Flattener
from cre.transform.feature_applier import FeatureApplier
from cre.transform.relative_encoder import RelativeEncoder
from cre.transform.vectorizer import Vectorizer



import sys, argparse

def parse_args(argv):
    parser = argparse.ArgumentParser(description='Utilities that help with cre usage')
    parser.add_argument('--clear-cache', dest='clear_cache', action='store_true',
         help="Clears the cre caches.")

    try:
        args = parser.parse_args(argv)
    except Exception:
        parser.print_usage()
        sys.exit()

    if(args.clear_cache):
        from cre_caching import clear_cache
        clear_cache()
    else:
        parser.print_usage()
        sys.exit()

def main():
    args = parse_args(sys.argv[1:])




from cre.context import cre_context
from cre.memset import MemSet

class MemSetBuilder():
    def __init__(self, out_memsets=None, context=None):
        self.context = cre_context(context)
        self.non_relational_attrs = {}
        self.relational_attrs = {}
        self.out_memsets = out_memsets

    def get_non_relational(self, type_name):
        if(type_name not in self.non_relational_attrs):
            fact_type = self.context.get_type(name=type_name)
            attrs = list(fact_type.filter_spec("~relational").keys())
            self.non_relational_attrs[type_name] = attrs
        return self.non_relational_attrs[type_name]

    def get_relational(self, type_name):
        if(type_name not in self.relational_attrs):
            fact_type = self.context.get_type(name=type_name)
            attrs = list(fact_type.filter_spec("relational").keys())
            self.relational_attrs[type_name] = attrs
        return self.relational_attrs[type_name]

    def tranform(self, state_dict, out_memset=None, return_map=False):
        '''Converts a dictionary of dictionaries each representing a fact into a MemSet'''
        # Make each fact instance, but skip setting any relational members.
        fact_instances = {}
        for _id, config in state_dict.items():
            fact_type = self.context.get_type(name=config['type'])
            non_rel_attrs = self.get_non_relational(config['type'])
            kwargs = {attr:config[attr] for attr in non_rel_attrs if attr in config} 
            fact = fact_type(**kwargs)
            fact_instances[_id] = fact

        # Now that all facts exist set any relational attributes
        for _id, fact in fact_instances.items():
            config = state_dict[_id]
            rel_attrs = self.get_relational(config['type'])

            for attr in rel_attrs:
                if(attr not in config): continue
                val_name = config[attr]
                if(not val_name): continue # i.e. skip if None or empty str
                if(val_name not in fact_instances):
                    raise ValueError(f"Reference to unspecified fact {val_name}.")
                setattr(fact, attr, fact_instances[val_name])

        # Declare each fact to a new MemSet
        if(out_memset is None):
            out_memset = self.out_memsets if(self.out_memsets) else MemSet(auto_clear_refs=True)
        for _id, fact in fact_instances.items():
            out_memset.declare(fact)

        if(return_map):
            return out_memset, fact_instances
        else:
            return out_memset

    def __call__(self, state_dict, *args, **kwargs):
        return self.tranform(state_dict, *args, **kwargs)



import numpy as np
import sys,os
import hashlib 
import timeit
from types import FunctionType

import os
import time
from pathlib import Path
import shutil
# from cre.caching import cache_dir

#Snatch AppDirs from numba numba and find the cache dir
from numba.misc.appdirs import AppDirs
from base64 import b64encode

#Resolve the location of the cache_dir
appdirs = AppDirs(appname="cre", appauthor=False)
cache_dir = os.path.join(appdirs.user_cache_dir,"cre_cache")
os.environ['NUMBA_CACHE_DIR'] = os.path.join(os.path.split(cache_dir)[0], "numba_cache")
print("CRE Cache Lives Here: ", cache_dir)

#Make the cache_dir if it doesn't exist
if not os.path.exists(cache_dir):
	os.makedirs(cache_dir)
# Path(os.path.join(cache_dir,"__init__.py")).touch(exist_ok=True)

#Add cre_cache to path
sys.path.insert(0, appdirs.user_cache_dir)

def clear_cache():
	#Clear the cache folder determined by AppDirs
	shutil.rmtree(os.path.abspath(os.path.join(cache_dir,"../")),ignore_errors=True)

	#Clear __pycache__ in /cre
	shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__),"../cre/__pycache__")), ignore_errors=True)
	print("Cache Cleared!")

import linecache
def cache_safe_exec(source,lcs=None,gbls=None,cache_name='cache-safe'):
    fp = "<ipython-%s>" %cache_name
    lines = [line + '\n' for line in source.splitlines()]
    linecache.cache[fp] = (len(source), None, lines, fp)
    code = compile(source,fp,'exec')
    l = lcs if lcs is not None else {}
    g = gbls if gbls is not None else globals()
    exec(code,g,l)
    return l,g


class _UniqueHashable():
	def get_hashable(self):
		raise NotImplemented()

	def get_hashcode(self):
		return unique_hash(self.get_hashable())
	


def update_unique_hash(m,obj):
	if(isinstance(obj,str)):
		m.update(obj.encode('utf-8'))
	elif(isinstance(obj,(tuple,list, np.ndarray))):
		for i,x in enumerate(obj):
			update_unique_hash(m,i)
			update_unique_hash(m,x)
	elif(isinstance(obj,dict)):
		for k,v in obj.items():
			update_unique_hash(m,k)
			update_unique_hash(m,v)
	elif(isinstance(obj,FunctionType)):
		m.update(obj.__code__.co_code)
	elif(isinstance(obj,_UniqueHashable)):
		update_unique_hash(m,obj.get_hashable())
	elif(isinstance(obj,bytes)):
		m.update(obj)
	else:
		m.update(str(obj).encode('utf-8'))


def unique_hash(stuff, hash_func='sha256'):
	'''Returns a 64-bit encoded hashstring of some 'stuff' '''
	m = hashlib.new(hash_func)
	update_unique_hash(m,stuff)	

	# Encode in base64 map the usual altchars '/' and "+' to 'A' and 'B".
	s = b64encode(m.digest(),altchars=b'AB').decode('utf-8')
	# Strip the trailing '='.
	s = s[:-1]
	return s

def unique_hash_v(stuff, hash_func='sha256'):
	'''Same as unique_hash but adds in cre.__version__'''
	from cre import __version__
	m = hashlib.new(hash_func)
	update_unique_hash(m,__version__)	
	update_unique_hash(m,stuff)	

	# Encode in base64 map the usual altchars '/' and "+' to 'A' and 'B".
	s = b64encode(m.digest(),altchars=b'AB').decode('utf-8')
	# Strip the trailing '='.
	s = s[:-1]
	return s


def get_cache_path(name,hsh=None,suffix=".py"):
	if(hsh is None):
		return os.path.join(cache_dir,name+suffix)
	else:
		return os.path.join(cache_dir,name,"_" + str(hsh) +suffix)

def source_in_cache(name,hsh=None):
	path = get_cache_path(name,hsh)
	return path if os.path.exists(path) else None


def source_from_cache(name,hsh):
	path = get_cache_path(name,hsh)
	with open(path,mode='r') as f:
		out = f.read()
	return out

def aot_compile(name,hsh):
	out = import_from_cached(name,hsh,['cc'])	
	out['cc'].compile()

def source_to_cache(name,hsh,source,is_aot=False):
	path = get_cache_path(name,hsh)
	os.makedirs(os.path.dirname(path), exist_ok=True)
	with open(path,mode='w') as f:
		f.write(source)
		f.flush()
	if(is_aot): aot_compile(name,hsh)


def gen_import_str(name,hsh,targets,aot_module=None):
	aot_module = aot_module if aot_module else ''
	if(isinstance(targets,dict)):
		targets_str = ", ".join([f'{k} as {v}' for k,v in targets.items()])
	else:
		targets_str = ", ".join(targets)
	return f"from cre_cache.{name}.{aot_module}_{hsh} import {targets_str}"


def _import_cached(name,hsh,aot_module=None):
	if(not aot_module): aot_module = ''
	mod_str = f'cre_cache.{name}.{aot_module}_{hsh}'
	try:
		return importlib.import_module(mod_str)
	except (ModuleNotFoundError) as e:
		#Invalidates any the finder caches in case the module has was newly created
		importlib.invalidate_caches()
		return importlib.import_module(mod_str)

import importlib
def import_from_cached(name,hsh,targets,aot_module=None):
	if(aot_module):
		try:
			#Try to import the AOT Module
			mod = _import_cached(name,hsh,aot_module)
		except (ImportError,AttributeError) as e:
			try:
				#If couldn't import try to compile the AOT Module
				aot_compile(name,hsh)
				mod = _import_cached(name,hsh,aot_module)
			except Exception as e:
				#If still couldn't import or recompile, import as cached
				mod = _import_cached(name,hsh)
	else:
		mod = _import_cached(name,hsh)
		
	return {x:getattr(mod,x) for x in targets}
	# l = {x:getattr(mod,x) for x in targets}
	# return {k:l[k] for k in targets}
	# path = get_cache_path(name,hsh)




if(__name__ == "__main__"):
	from numba import njit, jit
	from cre.operator import BaseOperator
	from cre.gensource import *
	import numpy as np
	import importlib

	class PoopyMeta(type, _UniqueHashable):
		def get_hashable(cls):
			d = {k: v for k,v in vars(cls).items() if not k in ['get_hashable','__weakref__','__doc__','__dict__']}
			print("WHEE",d)
			return d

	class POOPY(metaclass=PoopyMeta):
		crabs = 1
		def shloopy():
			return 1+2

		# @classmethod
		



	print(POOPY.get_hashable())
	# print(Add.get_hashable())

	def myfunc():
		pass

	print(myfunc.__code__.co_code)
	# myfunc.__code__.__text_signature__



	N=1000
	def time_ms(f):
		f() #warm start
		return " %0.6f ms" % (1000.0*(timeit.timeit(f, number=N)/float(N)))


	#Basic tests
	obj = {"A":1,"B":2,"C":3}
	obj2 = ["A",1,"B",2,"C",3]
	obj3 = ["A",myfunc,"B",2,"C",3]
	obj4 = ["A",POOPY.shloopy,"B",2,"C",3]
	obj4 = ["A",POOPY,"B",2,"C",3]
	# obj4 = ["A",Add,"B",2,"C",3]


	print(unique_hash(obj))
	print(unique_hash(obj2))
	print(unique_hash(obj3))
	print(unique_hash(obj4))

	source = '@njit'
	print(hash(source))

	hash_code = POOPY.get_hashcode()

	print(hash_code)

	print(get_cache_path('hello',hash_code))
	source_to_cache("hello", hash_code,source)

	print(source_in_cache("hello", hash_code))
	print(source_from_cache("hello", hash_code))

	#Start op test
	class Floop(BaseOperator):
		signature = 'float(float,float)'
		def condition(x, y):
			return x > y and y > 0.0
		def forward(x, y):
			return x + y

	hash_code = Floop.get_hashcode()
	source = gen_source_broadcast_forward(Floop,True)

	print(get_cache_path('Floop',hash_code))
	source_to_cache("Floop", hash_code,source)

	print(source_in_cache("Floop", hash_code))
	print(source_from_cache("Floop", hash_code))


	print(hash_code)	
	print(source)

	# f = njit(Floop.forward)
	# c = njit(Floop.condition)
	# g = {'f': f, "c" : c, "jit" : jit}
	# l = {}

	# cache_path = get_cache_path('Floop',hash_code)
	# spec = importlib.util.spec_from_file_location("cre.cache.type1", cache_path)
	# my_mod = importlib.util.module_from_spec(spec)
	# spec.loader.exec_module(my_mod)


	# from cre_cache.Floop._36e47400043b53379e884eddd923dff7aa3f52cc5ab38f3a3f42c5eaa01d942e import Floop_forward
	# exec("from cre_cache.Floop._36e47400043b53379e884eddd923dff7aa3f52cc5ab38f3a3f42c5eaa01d942e import Floop_forward", g, l)

	# l['Floop_forward'](np.array([1.0,2.0]))

	spec = {
		"x" : "number",
		"y" : "number",
	}
	source = gen_source_standard_imports()
	source += gen_source_tuple_defs("Point",spec)
	source += gen_source_get_enumerized("Point",spec)
	source += gen_source_enumerize_nb_objs("Point",spec)
	source += gen_source_pack_from_numpy("Point",spec)
	# print(source)
	
	hash_code = unique_hash(["Point",spec])
	source_to_cache("Point",hash_code,source)
	# print(hash_code)
	print(source_from_cache("Point", hash_code))
	d = import_from_cached("Point",hash_code,["Point"])
	# exec("from cre_cache.Point._b666f2d9dfe40302fff737e562dc999fb89e3cd4be6f2f0200875eadf7286613 import Point", {}, l)
	print(d['Point'])







